# dbdu.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""The Dbdu class for methods shared by Berkeley DB interface modules.

This module is relevant to the berkeleydb, bsddb3, and tcl, interfaces to
Berkeley DB.

"""
import os
import zipfile

from solentware_base.core.constants import (
    SUBFILE_DELIMITER,
    EXISTENCE_BITMAP_SUFFIX,
    SEGMENT_SUFFIX,
)

from ..core.filespec import (
    SECONDARY,
    DB_ENVIRONMENT_GIGABYTES,
    DB_ENVIRONMENT_BYTES,
    DB_ENVIRONMENT_MAXLOCKS,
)
from .archivedu import _delete_archive, _archive
from .alldu import get_filespec
from .dptcompatdu import DptCompatdu


class Dbdu(DptCompatdu):
    """Provide deferred update methods shared by the Berkeley DB interfaces.

    The methods provided by DptCompatdu are shared with engines other than
    Berkeley DB.

    The whole database can be put in a single file, or each table (called a
    database in Berkeley DB terminology) in the database can be put in a
    file of it's own.
    """

    def __init__(self, databasefile, exception_class, flags, **kargs):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        assert issubclass(exception_class, Exception)
        environment = {
            "flags": flags,
            "gbytes": DB_ENVIRONMENT_GIGABYTES,
            "bytes": DB_ENVIRONMENT_BYTES,
            "maxlocks": DB_ENVIRONMENT_MAXLOCKS,
        }

        try:
            names = get_filespec(**kargs)
        except Exception as error:
            if __name__ == "__main__":
                raise
            raise exception_class("DB description invalid") from error

        try:
            super().__init__(names, databasefile, environment, **kargs)
        except Exception as error:
            if __name__ == "__main__":
                raise
            raise exception_class("DB description invalid") from error

    @staticmethod
    def open_context_prepare_import():
        """Return True.

        No preparation actions thet need database open for Berkeley DB.

        """
        return True

    def get_archive_names(self, files=()):
        """Return specified files and existing operating system files."""
        if self.home_directory is None:
            return None, []
        if self._file_per_database:
            names = dict()
            for key in self.specification:
                if key not in files:
                    continue
                name_list = []
                names[os.path.join(self.home_directory, key)] = name_list
                for item in self.specification[key][SECONDARY]:
                    name_list.append(
                        os.path.join(
                            self.home_directory,
                            SUBFILE_DELIMITER.join((key, item)),
                        )
                    )
                name_list.append(
                    os.path.join(
                        self.home_directory,
                        SUBFILE_DELIMITER.join((key, EXISTENCE_BITMAP_SUFFIX)),
                    )
                )
                name_list.append(
                    os.path.join(
                        self.home_directory,
                        SUBFILE_DELIMITER.join((key, SEGMENT_SUFFIX)),
                    )
                )
            exists = [
                os.path.basename(n)
                for n in names
                if os.path.exists(".".join((n, "zip")))
            ]
            return names, exists
        names = [self.database_file]
        exists = [
            os.path.basename(n)
            for n in names
            if os.path.exists(".".join((n, "bz2")))
        ]
        return names, exists

    def archive(self, flag=None, names=None):
        """Write a bz2 or zip backup of files containing games.

        Intended to be a backup in case import fails.

        """
        if self.home_directory is None:
            return None
        if names is None:
            return False
        if not self.delete_archive(flag=flag, names=names):
            return None
        if flag:
            if self._file_per_database:
                for name in names:
                    archiveguard = ".".join((name, "grd"))
                    archivename = ".".join((name, "zip"))
                    with zipfile.ZipFile(
                        archivename,
                        mode="w",
                        compression=zipfile.ZIP_DEFLATED,
                        allowZip64=True,
                    ) as zip_archive:
                        for source in names[name]:
                            zip_archive.write(
                                source, arcname=os.path.basename(source)
                            )
                    with open(archiveguard, "wb"):
                        pass
            else:
                _archive(names)
        return True

    def delete_archive(self, flag=None, names=None):
        """Delete a zip backup of files containing games."""
        if self.home_directory is None:
            return None
        if names is None:
            return False
        if flag:
            if self._file_per_database:
                not_backups = []
                for name in names:
                    archiveguard = ".".join((name, "grd"))
                    archivename = ".".join((name, "zip"))
                    if not os.path.exists(archivename):
                        try:
                            os.remove(archiveguard)
                        except FileNotFoundError:
                            pass
                        continue
                    with zipfile.ZipFile(
                        archivename,
                        mode="r",
                        compression=zipfile.ZIP_DEFLATED,
                        allowZip64=True,
                    ) as zip_archive:
                        namelist = zip_archive.namelist()
                        extract = [
                            e
                            for e in namelist
                            if os.path.join(self.home_directory, e)
                            in names[name]
                        ]
                        if len(extract) != len(namelist):
                            not_backups.append(os.path.basename(archivename))
                            continue
                    try:
                        os.remove(archiveguard)
                    except FileNotFoundError:
                        pass
                    try:
                        os.remove(archivename)
                    except FileNotFoundError:
                        pass
                if not_backups:
                    return None
            else:
                _delete_archive(names)
        return True



# database.py
# Copyright 2019 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""ChessTab database methods common to all database engine interfaces."""

import os
import bz2
import shutil

from ..core.filespec import (
    NEWGAMES_FIELD_DEF,
    NEWGAMES_FIELD_VALUE,
    PARTIAL_FILE_DEF,
)
from .. import APPLICATION_NAME, ERROR_LOG


class Database:
    """Define methods which are common to all database engine interfaces."""

    def use_deferred_update_process(self, **kargs):
        """Return path to deferred update module.

        **kargs - soak up any arguments other database engines need.

        """
        del kargs
        return self._deferred_update_process

    def open_database(self, files=None):
        """Return True to fit behaviour of dpt version of this method."""
        super().open_database(files=files)
        return True

    @staticmethod
    def dump_database(names=()):
        """Dump database in compressed files."""
        for name in names:
            compressor = bz2.BZ2Compressor()
            archivename = ".".join((name, "broken", "bz2"))
            with open(name, "rb") as file_in, open(
                archivename, "wb"
            ) as file_out:
                inp = file_in.read(10000000)
                while inp:
                    compressed = compressor.compress(inp)
                    if compressed:
                        file_out.write(compressed)
                    inp = file_in.read(10000000)
                compressed = compressor.flush()
                if compressed:
                    file_out.write(compressed)

    @staticmethod
    def delete_backups(names=()):
        """Delete backup files."""
        for name in names:
            archiveguard = ".".join((name, "grd"))
            archivename = ".".join((name, "bz2"))
            try:
                os.remove(archiveguard)
            except FileNotFoundError:
                pass
            try:
                os.remove(archivename)
            except FileNotFoundError:
                pass

    @staticmethod
    def restore_backups(names=()):
        """Restore database from backup files."""
        for name in names:
            decompressor = bz2.BZ2Decompressor()
            archivename = ".".join((name, "bz2"))
            with open(archivename, "rb") as file_in, open(
                name, "wb"
            ) as file_out:
                inp = file_in.read(1000000)
                while inp:
                    decompressed = decompressor.decompress(inp)
                    if decompressed:
                        file_out.write(decompressed)
                    inp = file_in.read(1000000)
        return True

    @staticmethod
    def _delete_database_names():
        """Return tuple of filenames to delete from database directory.

        Subclasses should override this method to delete the relevant files.

        """
        return ()

    def delete_database(self):
        """Close and delete the open chess database."""
        listnames = set(n for n in os.listdir(self.home_directory))
        homenames = set(
            n
            for n in self._delete_database_names()
            if os.path.basename(n) in listnames
        )
        if ERROR_LOG in listnames:
            homenames.add(os.path.join(self.home_directory, ERROR_LOG))
        if len(listnames - set(os.path.basename(h) for h in homenames)):
            message = "".join(
                (
                    "There is at least one file or folder in\n\n",
                    self.home_directory,
                    "\n\nwhich may not be part of the database.  These items ",
                    "have not been deleted by ",
                    APPLICATION_NAME,
                    ".",
                )
            )
        else:
            message = None
        self.close_database()
        for pathname in homenames:
            if os.path.isdir(pathname):
                shutil.rmtree(pathname, ignore_errors=True)
            else:
                os.remove(pathname)
        try:
            os.rmdir(self.home_directory)
        except FileNotFoundError:
            pass
        return message

    def get_archive_names(self, files=()):
        """Return names and operating system files for archives and guards."""
        del files
        names = (self.database_file,)
        archives = dict()
        guards = dict()
        for name in names:
            archiveguard = ".".join((name, "grd"))
            archivefile = ".".join((name, "bz2"))
            for box, file in ((archives, archivefile), (guards, archiveguard)):
                if os.path.exists(file):
                    box[name] = file
        return (names, archives, guards)

    def open_after_import_without_backups(self, files=()):
        """Return True after doing database engine specific open actions.

        For SQLite3 and Berkeley DB just call open_database.

        """
        del files
        super().open_database()

        # Return True to fit behaviour of chessdpt module version of method.
        return True

    def open_after_import_with_backups(self, files=()):
        """Return True after doing database engine specific open actions.

        For SQLite3 and Berkeley DB just call open_database.

        """
        del files
        super().open_database()

        # Return True to fit behaviour of chessdpt module version of method.
        return True

    def save_broken_database_details(self, files=()):
        """Save database engine specific detail of broken files to be restored.

        It is assumed that the Database Services object exists.

        """

    def adjust_database_for_retry_import(self, files):
        """Database engine specific actions done before re-trying an import."""

    def mark_partial_positions_to_be_recalculated(self):
        """File all partial positions to be recalculated."""
        self.start_transaction()
        allrecords = self.recordlist_ebm(PARTIAL_FILE_DEF)
        self.file_records_under(
            PARTIAL_FILE_DEF,
            NEWGAMES_FIELD_DEF,
            allrecords,
            self.encode_record_selector(NEWGAMES_FIELD_VALUE),
        )
        allrecords.close()
        self.commit()



# alldu.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Functions, and Alldu class for methods, used by most database interfaces.

'Most' means all database interfaces except DPT.
"""
from solentware_base.core.segmentsize import SegmentSize
from solentware_base.core.constants import FILEDESC

from ..core.filespec import FileSpec
from ..core.chessrecord import ChessDBrecordGameImport


def chess_du(
    cdb, pgnpaths, file_records, reporter=lambda text, timestamp=True: None
):
    """Open database, import games and close database."""
    del file_records
    importer = ChessDBrecordGameImport()
    cdb.open_database()
    cdb.set_defer_update()
    for pgnfile in pgnpaths:
        with open(pgnfile, "r", encoding="iso-8859-1") as source:
            importer.import_pgn(cdb, source, pgnfile, reporter=reporter)
    if reporter is not None:
        reporter("Finishing import: please wait.")
        reporter("", timestamp=False)
    cdb.do_final_segment_deferred_updates()
    cdb.unset_defer_update()
    cdb.close_database()


def get_filespec(**kargs):
    """Return FileSpec instance with FILEDESCs removed at **kargs request.

    The FILEDESCs are deleted if allowcreate is False, the default.
    """
    names = FileSpec(**kargs)
    if not kargs.get("allowcreate", False):
        for table_name in names:
            if FILEDESC in names[table_name]:
                del names[table_name][FILEDESC]
    return names


class Alldu:
    """Provide deferred update methods shared by all interfaces.

    All the supported engines follow DPT in dividing the numeric primary
    keys into fixed-size segments.  When importing games a large amount of
    memory is required depending on number of games.  Some operating
    systems limit the memory available to a process.  The class attribute
    deferred_update_points is set when the segment size is greater than
    32768 in an attempt to avoid a MemoryError exception.
    """

    # The optimum chunk size is the segment size.
    # Assuming 2Gb memory:
    # A default FreeBSD user process will not cause a MemoryError exception for
    # segment sizes up to 65536 records, so the optimum chunk size defined in
    # the superclass will be selected.
    # A MS Windows XP process will cause the MemoryError exeption which selects
    # the 32768 game chunk size.
    # A default OpenBSD user process will cause the MemoryError exception which
    # selects the 16384 game chunk size.
    # The error log problem fixed at chesstab-0.41.9 obscured what was actually
    # happening: OpenBSD gives a MemoryError exception but MS Windows XP heads
    # for thrashing swap space in some runs with a 65536 chunk size (depending
    # on order of processing indexes I think). Windows 10 Task Manager display
    # made this obvious.
    # The MemoryError exception or swap space thrashing will likely not occur
    # for a default OpenBSD user process or a MS Windows XP process with
    # segment sizes up to 32768 records. Monitoring with Top and Task Manager
    # suggests it gets close with OpenBSD.

    # pylint comparison-with-callable report is false positive.
    # Perhaps because db_segment_size is a property and the last statement
    # in segmentsize module is 'SegmentSize = SegmentSize()'.
    if SegmentSize.db_segment_size > 32768:
        for f, m in ((4, 700000000), (2, 1400000000)):
            try:
                b" " * m
            except MemoryError:

                # Override the value in the superclass.
                deferred_update_points = frozenset(
                    i
                    for i in range(
                        65536 // f - 1,
                        SegmentSize.db_segment_size,
                        65536 // f,
                    )
                )

                break
        del f, m



# gamedbdelete.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise delete toplevel to delete chess game record."""

from solentware_grid.gui.datadelete import DataDelete

from pgn_read.core.parser import PGN
from pgn_read.core.constants import TAG_WHITE, TAG_BLACK

from .gametoplevel import GameToplevel
from .toplevelpgn import DeletePGNToplevel


class GameDbDelete(DeletePGNToplevel, DataDelete):
    """Delete PGN text for game from database.

    parent is used as the master argument in a GameToplevel call.

    ui is used as the ui argument in a GameToplevel call.

    parent, oldobject, and the GameToplevel instance created, are used as
    arguments in the super.__init__ call.

    Attribute pgn_score_name provides the name used in widget titles and
    message text.

    Methods _get_title_for_object and _set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    pgn_score_name = "Game"

    def __init__(self, parent=None, oldobject=None, ui=None):
        """Extend and create toplevel widget for deleting chess game."""
        # Toplevel title set '' in __init__ and to proper value in _initialize.
        super().__init__(
            instance=oldobject,
            parent=parent,
            oldview=GameToplevel(master=parent, ui=ui),
            title="",
        )
        self._initialize()

    @property
    def ui_base_table(self):
        """Return the User Interface TagRosterGrid object."""
        return self.ui.base_games

    @property
    def ui_items_in_toplevels(self):
        """Return the User Interface objects in Toplevels."""
        return self.ui.games_and_repertoires_in_toplevels

    @property
    def ui(self):
        """Return the User Interface object from 'read-only' view."""
        return self.oldview.ui

    def _set_item(self, view, object_):
        """Populate view with the repertoire extracted from object_."""
        self._set_default_source_for_object(object_)
        view.set_position_analysis_data_source()
        view.collected_game = next(
            PGN(game_class=view.gameclass).read_games(object_.get_srvalue())
        )
        view.set_and_tag_item_text()

    def _get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a Game object_.

        Default value of object_ is object attribute from DataDelete class.

        """
        if object_ is None:
            object_ = self.object
        try:
            tags = object_.value.collected_game.pgn_tags
            return "  ".join(
                (
                    self.pgn_score_name.join(("Delete ", ":")),
                    " - ".join((tags[TAG_WHITE], tags[TAG_BLACK])),
                )
            )
        except TypeError:
            return self.pgn_score_name.join(
                ("Delete ", " - names unknown or invalid")
            )
        except KeyError:
            return self.pgn_score_name.join(
                ("Delete ", " - names unknown or invalid")
            )

    def _set_default_source_for_object(self, object_=None):
        """Set default source for Toplevel containing a Game object_.

        Default value of object_ is object attribute from DataDelete class.

        Currently do nothing for games.  Originally used for games with PGN
        errors, where it was the name of the PGN file containing the game.

        Now present for compatibility with Repertoires.

        """

    # Resolve pylint message arguments-differ deferred.
    # Depends on detail of planned naming of methods as private if possible.
    # mark...recalculated starts and commits a transaction unconditionally.
    # No harm in using the same default as the 'super()' method.
    def delete(self, commit=True):
        """Mark partial position records for recalculation and return key."""
        self.datasource.dbhome.mark_partial_positions_to_be_recalculated()
        super().delete(commit=commit)



# chesstab.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Run ChessTab with available database engine highest in preference order.

For existing databases the choice is restricted to database engines able to
open the database.

When allowcreate evaluates not true, databases cannot be created.

"""

if __name__ == "__main__":

    from chesstab.gui.chess import Chess

    app = Chess(allowcreate=True)
    app.root.mainloop()



# dptcompatdu.py
# Copyright 2022 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""The DptCompatdu class for methods shared by non-DPT interface modules."""


class DptCompatdu:
    """Provide do nothing methods for compatibility with DPT interface."""

    def add_import_buttons(self, *a):
        """Return None.

        No extra buttons required.  Method exists for DPT compatibility.

        """

    # pylint no-self-use message.
    # Why this method and not the other two?
    # Does it imply 'self' should not be removed given purpose of methods?
    @staticmethod
    def get_file_sizes():
        """Return an empty dictionary.

        No sizes needed.  Method exists for DPT compatibility.

        """
        return dict()

    def report_plans_for_estimate(self, estimates, reporter):
        """Remind user to check estimated time to do import.

        No planning needed.  Method exists for DPT compatibility.

        """
        # See comment near end of class definition Chess in relative module
        # ..gui.chess for explanation of this change.
        # reporter.append_text_only(''.join(
        #    ('The expected duration of the import may make starting ',
        #     'it now inconvenient.',
        #     )))
        # reporter.append_text_only('')



# cqlnode.py
# Copyright 2017 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess Query Language (ChessQL) parameter and filter evaluator.

Evaluate a node tree built by cql package using a chesstab database.

The description of variables at CQL version 6.0.4 suggests any construct using
variables cannot be evaluated by ChessTab because indicies are used rather than
processing each game move-by-move.  The relevant passages are copied from CQL
documentation as comments in the chessql.core.statement module.

"""
import copy

from chessql.core.cql import Token
from chessql.core.constants import (
    RANGE_SEPARATOR,
    PIECE_NAMES,
    SQUARE_DESIGNATOR_SEPARATOR,
    FILE_NAMES,
    CQL_RANK_NAMES,
)
from chessql.core.node import Node
from chessql.core.piecedesignator import PieceDesignator

from ..basecore.rayfilter import RayFilter


class CQLNodeError(Exception):
    """Exception class for cqlnode module."""


class CQLNode(Node):
    """Extend ChessQL Node for ChessTab implementation of CQL statements.

    data holds the components used to build a statement which can be
    processed by the solentware_base.core where and find modules.

    """

    def __init__(self, *a, **k):
        """Delegate then initialize node state."""
        super().__init__(*a, **k)

        # May get rid of 'where' because components should not be needed again
        # after the 'where statement' has been constructed: bind the 'where
        # statement' to 'data' in other words.
        self.data = None
        self.where = None

    @staticmethod
    def transform_piece_designators(fs_filter):
        """Apply transforms to piece designators in CQL statement."""
        children = fs_filter.children
        for node in fs_filter.node.children:
            children.append(FSNode(node))
            transformation = node.transform_piece_designators(children[-1])
            if transformation:
                return transformation
        transformation = fs_filter.transform_descendant_piece_designators()
        if transformation:
            return transformation
        return None

    def expand_child_piece_designators(self):
        """Expand piece designators in child nodes.

        Expanded piece designators are made available in FSNode.designator_set,
        and cache for optimizing index access.

        """
        for node in self.children:
            node.expand_child_piece_designators()
        if self.leaf:
            if self.tokendef is Token.PIECE_DESIGNATOR:
                piece_designator = PieceDesignator(self.leaf)
                piece_designator.parse()
                piece_designator.expand_piece_designator()
                self.data = piece_designator

    # Beginning of attempt to generalize self.data to instantiate set filter.
    # Deal with ray filter in addition to piece designator.
    def expand_set_filters(self):
        """Expand set filters in child nodes.

        Expanded piece designators are made available in FSNode.designator_set,
        and cache for optimizing index access.

        """
        for node in self.children:
            node.expand_set_filters()
        if self.leaf:
            if self.tokendef is Token.PIECE_DESIGNATOR:
                piece_designator = PieceDesignator(self.leaf)
                piece_designator.parse()
                piece_designator.expand_piece_designator()
                self.data = piece_designator
            return
        if self.tokendef is Token.RAY:
            self.data = RayFilter(self)

    def get_shift_limits(self, ranklimits=None, filelimits=None):
        """Set rank and file shift limits."""
        if self.tokendef is Token.PIECE_DESIGNATOR:
            data = PieceDesignator(self.leaf)
            data.parse()
            data.get_shift_limits(ranklimits, filelimits)
        for node in self.children:
            node.get_shift_limits(ranklimits=ranklimits, filelimits=filelimits)

    def shift(self, shiftfiles, shiftranks):
        """Shift shiftfiles and shiftranks for pieces referenced."""
        if self.tokendef is Token.PIECE_DESIGNATOR:
            data = PieceDesignator(self.leaf)
            data.parse()
            squares = data.get_squares()
            squares = "".join([shiftfiles.get(s, s) for s in squares])
            squares = "".join([shiftranks.get(s, s) for s in squares])
            if data.is_compound_squares():
                squares = squares.join(("[", "]"))
            pieces = data.get_pieces()
            if data.is_compound_pieces():
                pieces = pieces.join(("[", "]"))
            self.leaf = pieces + squares
        for node in self.children:
            node.shift(shiftfiles, shiftranks)

    def rotate(self, rotation):
        """Rotate rotation for pieces referenced."""
        if self.tokendef is Token.PIECE_DESIGNATOR:
            data = PieceDesignator(self.leaf)
            data.parse()
            squares = _normalize_rotated_squares(
                "".join([rotation.get(s, s) for s in data.get_squares()])
            )
            if data.is_compound_squares():
                squares = squares.join(("[", "]"))
            pieces = data.get_pieces()
            if data.is_compound_pieces():
                pieces = pieces.join(("[", "]"))
            self.leaf = pieces + squares
        for node in self.children:
            node.rotate(rotation)

    def reflect_horizontal(self):
        """Reflect ranks for pieces referenced."""
        if self.tokendef is Token.PIECE_DESIGNATOR:
            data = PieceDesignator(self.leaf)
            data.parse()
            squares = _normalize_horizontally_reflected_squares(
                "".join(
                    [
                        FSNode.REFLECT_HORIZONTAL.get(s, s)
                        for s in data.get_squares()
                    ]
                )
            )
            if data.is_compound_squares():
                squares = squares.join(("[", "]"))
            pieces = data.get_pieces()
            if data.is_compound_pieces():
                pieces = pieces.join(("[", "]"))
            self.leaf = pieces + squares
        for node in self.children:
            node.reflect_horizontal()

    def reflect_vertical(self):
        """Reflect files for pieces referenced."""
        if self.tokendef is Token.PIECE_DESIGNATOR:
            data = PieceDesignator(self.leaf)
            data.parse()
            squares = _normalize_vertically_reflected_squares(
                "".join(
                    [
                        FSNode.REFLECT_VERTICAL.get(s, s)
                        for s in data.get_squares()
                    ]
                )
            )
            if data.is_compound_squares():
                squares = squares.join(("[", "]"))
            pieces = data.get_pieces()
            if data.is_compound_pieces():
                pieces = pieces.join(("[", "]"))
            self.leaf = pieces + squares
        for node in self.children:
            node.reflect_vertical()

    def rotate_and_reflect_horizontal(self):
        """Rotate 90 and reflect ranks for pieces referenced."""
        if self.tokendef is Token.PIECE_DESIGNATOR:
            data = PieceDesignator(self.leaf)
            data.parse()
            squares = _normalize_rotated_squares(
                "".join(
                    [
                        FSNode.ROTATE_90_REFLECT_HORIZONTAL.get(s, s)
                        for s in data.get_squares()
                    ]
                )
            )
            if data.is_compound_squares():
                squares = squares.join(("[", "]"))
            pieces = data.get_pieces()
            if data.is_compound_pieces():
                pieces = pieces.join(("[", "]"))
            self.leaf = pieces + squares
        for node in self.children:
            node.rotate_and_reflect_horizontal()

    def rotate_and_reflect_vertical(self):
        """Rotate 90 and reflect files for pieces referenced."""
        if self.tokendef is Token.PIECE_DESIGNATOR:
            data = PieceDesignator(self.leaf)
            data.parse()
            squares = _normalize_rotated_squares(
                "".join(
                    [
                        FSNode.ROTATE_90_REFLECT_VERTICAL.get(s, s)
                        for s in data.get_squares()
                    ]
                )
            )
            if data.is_compound_squares():
                squares = squares.join(("[", "]"))
            pieces = data.get_pieces()
            if data.is_compound_pieces():
                pieces = pieces.join(("[", "]"))
            self.leaf = pieces + squares
        for node in self.children:
            node.rotate_and_reflect_vertical()

    def flip_color(self):
        """Flip the colour of pieces referenced."""
        if self.tokendef is Token.PIECE_DESIGNATOR:
            data = PieceDesignator(self.leaf)
            data.parse()
            squares = _normalize_horizontally_reflected_squares(
                "".join(
                    [
                        FSNode.REFLECT_HORIZONTAL.get(s, s)
                        for s in data.get_squares()
                    ]
                )
            )
            if data.is_compound_squares():
                squares = squares.join(("[", "]"))
            pieces = "".join(
                FSNode.FLIP_COLOR_PIECE.get(s, s) for s in data.get_pieces()
            )
            if data.is_compound_pieces():
                pieces = pieces.join(("[", "]"))
            self.leaf = pieces + squares
        elif self.name == "plain_filter":
            if self.leaf in FSNode.FLIP_COLOR_TOMOVE:
                self.leaf = FSNode.FLIP_COLOR_TOMOVE[self.leaf]
        elif self.name in FSNode.FLIP_COLOR_FILTER:
            self.name = FSNode.FLIP_COLOR_FILTER[self.name]
            return
        for node in self.children:
            node.flip_color()

    def __deepcopy__(self, memo):
        """Return a deepcopy of memo."""
        newcopy = super().__deepcopy__(memo)
        newcopy.data = None
        newcopy.where = None
        return newcopy


class FSNode:
    """Scaffolding to evaluate CQLNodes created by the cql package."""

    INITIAL_RANK_LIMITS = CQL_RANK_NAMES[-1], CQL_RANK_NAMES[0]
    INITIAL_FILE_LIMITS = FILE_NAMES[-1], FILE_NAMES[0]

    ROTATE_90 = dict(
        zip(
            CQL_RANK_NAMES + FILE_NAMES,
            (
                "".join(z for z in reversed(FILE_NAMES))
                + "".join(z for z in CQL_RANK_NAMES)
            ),
        )
    )

    # This comprehension construct did not work at Python3.6.1 at this place.
    # ROTATE_180 = {x:ROTATE_90[ROTATE_90[x]]
    # for x in CQL_RANK_NAMES + FILE_NAMES}

    ROTATE_180 = {}
    for x in CQL_RANK_NAMES + FILE_NAMES:
        ROTATE_180[x] = ROTATE_90[ROTATE_90[x]]
    ROTATE_270 = {}
    for x in CQL_RANK_NAMES + FILE_NAMES:
        ROTATE_270[x] = ROTATE_90[ROTATE_180[x]]
    REFLECT_HORIZONTAL = dict(
        zip(CQL_RANK_NAMES, "".join(z for z in reversed(CQL_RANK_NAMES)))
    )
    REFLECT_VERTICAL = dict(
        zip(FILE_NAMES, "".join(z for z in reversed(FILE_NAMES)))
    )

    # These rotate-reflect combinations are not defined by CQL but are used to
    # generate the diagonal reflections needed to complete the flip transform.
    ROTATE_90_REFLECT_HORIZONTAL = {}
    ROTATE_90_REFLECT_VERTICAL = {}
    for x in CQL_RANK_NAMES + FILE_NAMES:
        y = ROTATE_90[x]
        ROTATE_90_REFLECT_HORIZONTAL[x] = REFLECT_HORIZONTAL.get(y, y)
        ROTATE_90_REFLECT_VERTICAL[x] = REFLECT_VERTICAL.get(y, y)
    del x, y

    FLIP_COLOR_PIECE = dict(
        zip(
            PIECE_NAMES,
            "".join(
                (
                    PIECE_NAMES[6:12],
                    PIECE_NAMES[0:6],
                    PIECE_NAMES[13],
                    PIECE_NAMES[12],
                    PIECE_NAMES[14],
                )
            ),
        )
    )
    FLIP_COLOR_TOMOVE = {
        Token.WTM.name: Token.BTM.name,
        Token.BTM.name: Token.WTM.name,
    }
    FLIP_COLOR_FILTER = {
        Token.WHITE.name: Token.BLACK.name,
        Token.BLACK.name: Token.WHITE.name,
    }

    def __init__(self, node):
        """Note node and initialize child nodes to empty list."""
        self.node = node
        self.children = []

    # Put this stuff in cql.core.statement because the piece designator text is
    # modified, not the expansion into database keys?
    def transform_descendant_piece_designators(self):
        """Apply transform to descendant piece designators.

        Some points from http://www.gadycosteff.com/cql/doc/transform.html for
        cql5.0 are:

        'flipvertical Pa1-8' is equivalent to 'Ph1-8 or Pa1-8'.  Replacing a
        single piece designator by a sequence of piece designators in an 'or'
        clause is valid for all the transform filters.

        The transform filters are: flipvertical, fliphorizontal, flipdihedral,
        flip, rotate90, rotate45, flipcolor, shifthorizontal, shiftvertical,
        and shift.  They are applied recursively to constituent filters:

        'flipvertical attack (Rg6 K)' and 'attack (Rg6 k) or attack (Rb6 k)'
        are equivalent.

        Each transform filter is in one of the categories: dihedral, rotate45,
        color, and shift.

        Ranges in transform filters specify the count of the transforms of the
        argument filter which must match for the transform filter to match.

        Comments on shift copied temporarely in full:

        shiftvertical shifts its argument 0 or more squares vertically:

        shiftvertical g6
        ≡ g1 or g2 or... or g8
        ≡ g1-8

        Likewise:

        shiftvertical [g2,g4]
        ≡ g1-8

        When a square is shifted off the board it normally disappears. Piece
        designators with empty square sets eliminate the entire transform:

        shiftvertical {Kb1 kg6}
        ≡
        {Kb1 Kg6} or {Kb2 Kg7} or {Kb3 Kg8}

        There is no downward shift of kg6 because doing so would eliminate the
        b1 square for the K.

        shifthorizontal works likewise.

        shift is equivalent to shifthorizontal shiftvertical. Using the example
        above:

        shift {Kb1 kg6}
        ≡ shifthorizontal shiftvertical {Kb1 kg6}
        ≡ shifthorizontal {Kb1 kg6} or
          shifthorizontal {Kb2 kg7} or
          shifthorizontal {Kb3 kg8}
        ≡ {Kb1 kg6} or {Ka1 kf6} or {Kc1 kh6} or
          {Kb2 kg7} or {Ka2 kf7} or {Kc2 kh7} or
          {Kb3 kg8} or {Ka3 kf8} or {Kc3 kh8}

        This also means that

        shift Ka2 ≡ K

        wraparound As a special rule, shiftvertical does not alter a file of 8
        squares:

        shiftvertical {Kd1-8 Ba2}
        ≡ {Kd1-8 Ba2} or
          {Kd1-8 Ba3} or
          {Kd1-8 Ba4} or
          {Kd1-8 Ba5} or
          {Kd1-8 Ba6} or
          {Kd1-8 Ba7} or
          {Kd1-8 Ba8} or
          {Kd1-8 Ba1}

        which turns out to be equivalent to

        {Kd1-8 Ba1-8}

        But

        shiftvertical {Kd2-8 Ba2}
        ≡ {Kd2-8 Ba2} or
          {Kd3-8 Ba3} or
          {Kd4-8 Ba4} or
          {Kd5-8 Ba5} or
          {Kd6-8 Ba6} or
          {Kd7-8 Ba7} or
          {Kd8 Ba8} or
          {Kd1-7 Ba1}

        which is entirely different. Similarly, shifthorizontal does not change
        ranks with 8 squares:

        shifthorizontal {Ka-h2 Ba4}
        ≡ {Ka-h2 Ba-h4}

        The shift transform doesn't change full ranks or files in the direction
        of its shift:

        shift {Ka-h2 Ba4}
        ≡ {Ka-h2 Ba-h4} or
          {Ka-h3 Ba-h5} or
          {Ka-h4 Ba-h6} or
          {Ka-h5 Ba-h7} or
          {Ka-h6 Ba-h8} or
          {Ka-h1 Ba-h3}

        Note that any transform applied to a piece designator without an
        explicit square qualifier leaves the piece designator unchanged:

        shift K
        ≡ K

        and likewise

        flip K
        ≡ K

        """
        for node in self.children:
            transformed = self._transform.get(
                node.node.tokendef, lambda s: None
            )(node)
            if transformed:
                return transformed
        return None

    def _flip(self):
        ranklow, rankhigh, filelow, filehigh = self._get_transform_limits()
        if (ranklow, rankhigh) == FSNode.INITIAL_RANK_LIMITS and (
            filelow,
            filehigh,
        ) == FSNode.INITIAL_FILE_LIMITS:
            return
        transforms = []
        for transformed in (
            FSNode.ROTATE_90,
            FSNode.ROTATE_180,
            FSNode.ROTATE_270,
        ):
            transforms.extend(self._generate_rotated_filters(transformed))
        transforms.extend(self._generate_vertical_reflection_filters())
        transforms.extend(self._generate_horizontal_reflection_filters())
        transforms.extend(
            self._generate_rotate_90_vertical_reflection_filters()
        )
        transforms.extend(
            self._generate_rotate_90_horizontal_reflection_filters()
        )
        for tnode in transforms:
            self.node.children.append(tnode.node)

    def _fliphorizontal(self):
        ranklow, rankhigh, filelow, filehigh = self._get_transform_limits()
        if (ranklow, rankhigh) == FSNode.INITIAL_RANK_LIMITS and (
            filelow,
            filehigh,
        ) == FSNode.INITIAL_FILE_LIMITS:
            return
        transforms = self._generate_horizontal_reflection_filters()
        for tnode in transforms:
            self.node.children.append(tnode.node)

    def _flipvertical(self):
        ranklow, rankhigh, filelow, filehigh = self._get_transform_limits()
        if (ranklow, rankhigh) == FSNode.INITIAL_RANK_LIMITS and (
            filelow,
            filehigh,
        ) == FSNode.INITIAL_FILE_LIMITS:
            return
        transforms = self._generate_vertical_reflection_filters()
        for tnode in transforms:
            self.node.children.append(tnode.node)

    def _rotate90(self):
        ranklow, rankhigh, filelow, filehigh = self._get_transform_limits()
        if (ranklow, rankhigh) == FSNode.INITIAL_RANK_LIMITS and (
            filelow,
            filehigh,
        ) == FSNode.INITIAL_FILE_LIMITS:
            return
        transforms = []
        for transformed in (
            FSNode.ROTATE_90,
            FSNode.ROTATE_180,
            FSNode.ROTATE_270,
        ):
            transforms.extend(self._generate_rotated_filters(transformed))
        for tnode in transforms:
            self.node.children.append(tnode.node)

    def _rotate45(self):
        ranklow, rankhigh, filelow, filehigh = self._get_transform_limits()
        if (ranklow, rankhigh) != FSNode.INITIAL_RANK_LIMITS or (
            filelow,
            filehigh,
        ) != FSNode.INITIAL_FILE_LIMITS:
            return "rotate45 on specific squares"
        return "rotate45 not implemented"

    def _flipcolor(self):
        """Apply flipcolor transform to filter.

        Some points from http://www.gadycosteff.com/cql/doc/transform.html for
        cql5.0 are:

        The flipcolor transform applied to filter is the 'or' of filter with
        the new filter formed from the filter as follows:

            the colors of any piece designators in filter are changed;
            'wtm' is changed to 'btm' and vice versa;
            'player white' is changed to 'player black' and vice versa;
            'elo white' is changed to 'elo black' and vice versa
            'result 0-1' is changed to 'result 1-0' and vice versa
            All piece designators in filter are reflected about the horizontal
            bisector of the board.

        """
        transforms = self._generate_flipped_color_filters()
        for tnode in transforms:
            self.node.children.append(tnode.node)

    def _shift(self):
        ranklow, rankhigh, filelow, filehigh = self._get_transform_limits()
        if (ranklow, rankhigh) == FSNode.INITIAL_RANK_LIMITS and (
            filelow,
            filehigh,
        ) == FSNode.INITIAL_FILE_LIMITS:
            # No shifts needed.
            # Change self.node.tokendef from 'shift' to '{' or leave alone so
            # meaning of self.node.range is clear?
            # Evaluation of query must treat this node as '{' rather than 'or'.
            # Is single child sufficient to decide this?
            # Same applies to all the other transform filters.
            return
        sourcefiles = list(FILE_NAMES)
        fileshifts = []
        limit_range = (
            sourcefiles.index(filelow) + 8 - sourcefiles.index(filehigh)
        )
        for count in range(sourcefiles.index(filelow)):
            sourcefiles.append(sourcefiles.pop(0))
        for count in range(limit_range):
            fileshifts.append(dict(zip(sourcefiles, FILE_NAMES)))
            sourcefiles.insert(0, sourcefiles.pop())
        # pylint message unused-variable.
        # Find a way of doing without count in this method's loops.
        del count
        sourceranks = list(CQL_RANK_NAMES)
        rankrange = (
            sourceranks.index(ranklow) + 8 - sourceranks.index(rankhigh)
        )
        for count in range(sourceranks.index(ranklow)):
            sourceranks.append(sourceranks.pop(0))
        transforms = []
        for count in range(rankrange):
            rankshifts = dict(zip(sourceranks, CQL_RANK_NAMES))
            for fshift in fileshifts:
                if (
                    fshift[filelow] != filelow
                    or rankshifts[ranklow] != ranklow
                ):
                    transforms.extend(
                        self._generate_shifted_filters(fshift, rankshifts)
                    )
            sourceranks.insert(0, sourceranks.pop())
        for tnode in transforms:
            self.node.children.append(tnode.node)

    def _shiftvertical(self):
        ranklimits = list(FSNode.INITIAL_RANK_LIMITS)
        self.node.get_shift_limits(ranklimits=ranklimits)
        if tuple(ranklimits) == FSNode.INITIAL_RANK_LIMITS:
            return
        self._shift_one_direction(ranklimits, CQL_RANK_NAMES, FILE_NAMES)

    def _shifthorizontal(self):
        filelimits = list(FSNode.INITIAL_FILE_LIMITS)
        self.node.get_shift_limits(filelimits=filelimits)
        if tuple(filelimits) == FSNode.INITIAL_FILE_LIMITS:
            return
        self._shift_one_direction(filelimits, FILE_NAMES, CQL_RANK_NAMES)

    _transform = {
        Token.FLIP: _flip,
        Token.FLIPHORIZONTAL: _fliphorizontal,
        Token.FLIPVERTICAL: _flipvertical,
        Token.ROTATE90: _rotate90,
        Token.ROTATE45: _rotate45,
        Token.FLIPCOLOR: _flipcolor,
        Token.SHIFT: _shift,
        Token.SHIFTHORIZONTAL: _shifthorizontal,
        Token.SHIFTVERTICAL: _shiftvertical,
    }

    def _get_transform_limits(self):
        ranklimit = list(FSNode.INITIAL_RANK_LIMITS)
        filelimit = list(FSNode.INITIAL_FILE_LIMITS)
        self.node.get_shift_limits(ranklimits=ranklimit, filelimits=filelimit)
        return ranklimit[0], ranklimit[1], filelimit[0], filelimit[1]

    def _shift_one_direction(self, limits, shiftsource, staticsource):
        """Extend children with one-direction transformed filters.

        Caller has specified transformation rules in a baseline square, limits,
        and rank and square shifts, shiftsource and staticsource depending on
        transform direction.

        self will be a transformation filter node so the transformed filters
        are appended to self.children.

        """
        source = list(shiftsource)
        shifts = []
        limit_range = source.index(limits[0]) + 8 - source.index(limits[1])
        for count in range(source.index(limits[0])):
            source.append(source.pop(0))
        for count in range(limit_range):
            shifts.append(dict(zip(source, shiftsource)))
            source.insert(0, source.pop())
        # pylint message unused-variable.
        # Find a way of doing without count in this method's loops.
        del count
        static = dict(zip(staticsource, staticsource))
        transforms = []
        for tshift in shifts:
            if tshift[limits[0]] != limits[0]:
                transforms.extend(
                    self._generate_shifted_filters(tshift, static)
                )
        for tnode in transforms:
            self.node.children.append(tnode.node)

    def _generate_shifted_filters(self, shiftfiles, shiftranks):
        transforms = []
        for node in self.children:
            transforms.append(copy.deepcopy(node))
            transforms[-1].node.shift(shiftfiles, shiftranks)
        return transforms

    def _generate_rotated_filters(self, rotation):
        transforms = []
        for node in self.children:
            transforms.append(copy.deepcopy(node))
            transforms[-1].node.rotate(rotation)
        return transforms

    def _generate_horizontal_reflection_filters(self):
        transforms = []
        for node in self.children:
            transforms.append(copy.deepcopy(node))
            transforms[-1].node.reflect_horizontal()
        return transforms

    def _generate_vertical_reflection_filters(self):
        transforms = []
        for node in self.children:
            transforms.append(copy.deepcopy(node))
            transforms[-1].node.reflect_vertical()
        return transforms

    def _generate_rotate_90_horizontal_reflection_filters(self):
        transforms = []
        for node in self.children:
            transforms.append(copy.deepcopy(node))
            transforms[-1].node.rotate_and_reflect_horizontal()
        return transforms

    def _generate_rotate_90_vertical_reflection_filters(self):
        transforms = []
        for node in self.children:
            transforms.append(copy.deepcopy(node))
            transforms[-1].node.rotate_and_reflect_vertical()
        return transforms

    def _generate_flipped_color_filters(self):
        transforms = []
        for node in self.children:
            transforms.append(copy.deepcopy(node))
            transforms[-1].node.flip_color()
        return transforms


def _normalize_rotated_squares(squares):
    normalized_squares = []
    for designator in squares.split(SQUARE_DESIGNATOR_SEPARATOR):
        square = list(designator)
        if len(square) == 2:
            if square[0] in CQL_RANK_NAMES:
                square[0], square[1] = square[1], square[0]
        elif len(square) == 6:
            if square[0] > square[2]:
                square[0], square[2] = square[2], square[0]
            if square[3] > square[5]:
                square[3], square[5] = square[5], square[3]
            if square[0] in CQL_RANK_NAMES:
                square[0], square[2], square[3], square[5] = (
                    square[3],
                    square[5],
                    square[0],
                    square[2],
                )
        elif square[1] == RANGE_SEPARATOR:
            if square[0] > square[2]:
                square[0], square[2] = square[2], square[0]
            if square[0] in CQL_RANK_NAMES:
                square.insert(0, square.pop())
        elif square[2] == RANGE_SEPARATOR:
            if square[1] > square[3]:
                square[1], square[3] = square[3], square[1]
            if square[0] in CQL_RANK_NAMES:
                square.append(square.pop(0))
        normalized_squares.append("".join(square))
    return SQUARE_DESIGNATOR_SEPARATOR.join(normalized_squares)


def _normalize_horizontally_reflected_squares(squares):
    normalized_squares = []
    for square in squares.split(SQUARE_DESIGNATOR_SEPARATOR):
        if (
            len(square) == 6
            or len(square) == 4
            and square[2] == RANGE_SEPARATOR
        ):
            square = list(square)
            square[-3], square[-1] = square[-1], square[-3]
            square = "".join(square)
        normalized_squares.append("".join(square))
    return SQUARE_DESIGNATOR_SEPARATOR.join(normalized_squares)


def _normalize_vertically_reflected_squares(squares):
    normalized_squares = []
    for square in squares.split(SQUARE_DESIGNATOR_SEPARATOR):
        if (
            len(square) == 6
            or len(square) == 4
            and square[1] == RANGE_SEPARATOR
        ):
            square = list(square)
            square[0], square[2] = square[2], square[0]
            square = "".join(square)
        normalized_squares.append("".join(square))
    return SQUARE_DESIGNATOR_SEPARATOR.join(normalized_squares)



# querydbshow.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise show toplevel to show game selection rule record."""

from solentware_grid.gui.datashow import DataShow

from .querytoplevel import QueryToplevel
from .topleveltext import ShowText


class QueryDbShow(ShowText, DataShow):
    """Show a game selection rule from database.

    parent is used as the master argument in a QueryToplevel call.

    ui is used as the ui argument in a QueryToplevel call.

    parent, oldobject, and the QueryToplevel instance created, are used as
    arguments in the super.__init__ call.

    Attribute text_name provides the name used in widget titles and message
    text.

    Methods _get_title_for_object and _set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    text_name = "Selection Rule Statement"

    def __init__(self, parent=None, oldobject=None, ui=None):
        """Extend and create toplevel widget to display game selection rule."""
        # Toplevel title set '' in __init__ and to proper value in _initialize.
        super().__init__(
            instance=oldobject,
            parent=parent,
            oldview=QueryToplevel(master=parent, ui=ui),
            title="",
        )
        if ui is not None:
            self.oldview.query_statement.set_database(
                ui.base_games.datasource.dbhome
            )
            self.oldview.query_statement.dbset = ui.base_games.datasource.dbset
        self._initialize()

    def _get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a selection rule object_.

        Default value of object_ is object attribute from DataShow class.

        """
        if object_ is None:
            object_ = self.object
        return "  ".join(
            (
                self.text_name.join(("Show ", ":")),
                object_.value.get_name_text(),
            )
        )

    @property
    def ui_base_table(self):
        """Return the User Interface QueryGrid object."""
        return self.ui.base_selections

    @property
    def ui_items_in_toplevels(self):
        """Return the User Interface objects in Toplevels."""
        return self.ui.selections_in_toplevels

    @property
    def ui(self):
        """Return the User Interface object from 'read-only' view."""
        return self.oldview.ui

    @staticmethod
    def _set_item(view, object_):
        """Populate view with the query extracted from object_."""
        view.query_statement.process_query_statement(object_.get_srvalue())
        view.set_and_tag_item_text()



# dptdu_dir.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import directory of PGN files with dpt.chessdptdu to database."""


if __name__ == "__main__":

    from .directory_widget import DirectoryWidget
    from ..dpt.chessdptdu import chess_dptdu

    DirectoryWidget(chess_dptdu, "dpt")



# gamedisplay.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Widgets to display and edit game scores.

These four classes display PGN text for games in the main window: they are
used in the gamelistgrid module.

The _GameDisplay class provides attributes and behaviour shared by the
GameDisplay, GameDisplayInsert, and GameDisplayEdit, classes.  It also provides
properties to support implementation of behaviour shared with the CQL*,
Repertoire*, and Query*, classes.

The GameDisplay, GameDisplayInsert, and GameDisplayEdit, classes are subclasses
of the relevant ShowPGN, InsertPGN, EditPGN, and DisplayPGN, classes from the
displaypgn module; to implement behaviour shared with all text widgets in the
main display (that includes widgets displaying text).

"""

from solentware_grid.core.dataclient import DataNotify

from solentware_bind.gui.bindings import Bindings

from .game import Game
from .gameedit import GameEdit
from ..core.chessrecord import ChessDBrecordGameUpdate
from .constants import STATUS_SEVEN_TAG_ROSTER_PLAYERS
from .eventspec import EventSpec
from .display import Display
from .displaypgn import ShowPGN, InsertPGN, EditPGN, DisplayPGN


class _GameDisplay(ShowPGN, DisplayPGN, Game, Display, Bindings, DataNotify):
    """Extend and link PGN game text to database.

    sourceobject - link to database.

    Attribute binding_labels specifies the order navigation bindings appear
    in popup menus.

    Attribute pgn_score_name provides the name used in widget titles and
    message text.

    Attribute pgn_score_tags provides the PGN tag names used in widget titles
    and message text.  It is the black, result, and white, PGN tags.

    Attribute pgn_score_source_name provides the error key value to index a
    PGN game score with errors.

    Attribute pgn_score_updater provides the class used to process PGN text
    into a database update.

    """

    binding_labels = (
        EventSpec.navigate_to_position_grid,
        EventSpec.gamedisplay_to_previous_game,
        EventSpec.analysis_to_scoresheet,
        EventSpec.gamedisplay_to_next_game,
        EventSpec.navigate_to_game_grid,
        EventSpec.scoresheet_to_analysis,
        EventSpec.navigate_to_repertoire_grid,
        EventSpec.navigate_to_active_repertoire,
        EventSpec.navigate_to_repertoire_game_grid,
        EventSpec.navigate_to_partial_grid,
        EventSpec.navigate_to_active_partial,
        EventSpec.navigate_to_partial_game_grid,
        EventSpec.navigate_to_selection_rule_grid,
        EventSpec.navigate_to_active_selection_rule,
        EventSpec.tab_traverse_backward,
        EventSpec.tab_traverse_forward,
    )

    # These exist so the insert_game_database methods in _GameDisplay and
    # repertoiredisplay._RepertoireDisplay, and delete_game_database in
    # GameDisplay and repertoiredisplay.RepertoireDisplay, can be modified and
    # replaced by single copies in the displaypgn.ShowPGN class.
    # See mark_partial_positions_to_be_recalculated() method too.
    # The names need to be more generic to make sense in cql, engine, and
    # query, context.
    pgn_score_name = "game"
    pgn_score_source_name = "Editor"
    pgn_score_tags = STATUS_SEVEN_TAG_ROSTER_PLAYERS
    pgn_score_updater = ChessDBrecordGameUpdate

    def __init__(self, sourceobject=None, **ka):
        """Extend and link PGN text of game to database."""
        super().__init__(**ka)
        self.blockchange = False
        if self.ui.base_games.datasource:
            self.set_data_source(self.ui.base_games.get_data_source())
        self.sourceobject = sourceobject

    # Could be put in game.Game class, but score.Score seems too deep.
    # Here can be justified because purpose is allow some methods to be moved
    # to displaypgn.ShowPGN class.
    @property
    def ui_displayed_items(self):
        """Return the displayed games."""
        return self.ui.game_items

    # Defined so _cycle_item and give_focus_to_widget methods can be shared by
    # gamedisplay._GameDisplay and repertoiredisplay._RepertoireDisplay
    # classes.
    @property
    def ui_configure_item_list_grid(self):
        """Return method to configure game grid widget."""
        return self.ui.configure_game_grid

    # ui_base_table and mark_partial_positions_to_be_recalculated defined
    # so insert_game_database method can be shared by gamedisplay._GameDisplay
    # and repertoiredisplay._RepertoireDisplay classes.
    # See class attributes pgn_score_name and pgn_score_source_name too.

    @property
    def ui_base_table(self):
        """Return grid widget for database games."""
        return self.ui.base_games

    @property
    def ui_items_in_toplevels(self):
        """Return game and repertoire items in Toplevels."""
        return self.ui.games_and_repertoires_in_toplevels

    @staticmethod
    def mark_partial_positions_to_be_recalculated(datasource=None):
        """Mark ChessQL statements, in datasource, to be recalculated."""
        datasource.dbhome.mark_partial_positions_to_be_recalculated()

    def _get_navigation_events(self):
        """Return event description tuple for navigation from game."""
        return (
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self.set_focus_repertoirepanel_item,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_partial,
                self.set_focus_partialpanel_item,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self.set_focus_selectionpanel_item,
            ),
            (EventSpec.gamedisplay_to_previous_game, self._prior_item),
            (EventSpec.gamedisplay_to_next_game, self._next_item),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            # No traverse_round because Alt-F8 toggles game and analysis.
        )

    def delete_item_view(self, event=None):
        """Remove game item from screen."""
        del event
        self.set_data_source()
        self.ui.delete_game_view(self)

    def on_game_change(self, instance):
        """Prevent update from self if instance refers to same record."""
        if self.sourceobject is not None:
            if (
                instance.key == self.sourceobject.key
                and self.datasource.dbname == self.sourceobject.dbname
                and self.datasource.dbset == self.sourceobject.dbset
            ):
                self.blockchange = True
            self._patch_pgn_score_to_fit_record_change_and_refresh_grid(
                self.ui.game_games, instance
            )

    def generate_popup_navigation_maps(self):
        """Return tuple of widget navigation map and switch to analysis map."""
        navigation_map = dict(self._get_navigation_events())
        local_map = {
            EventSpec.scoresheet_to_analysis: self.analysis_current_item,
        }
        return navigation_map, local_map


class GameDisplay(_GameDisplay, Game, DataNotify):
    """Display a chess game from a database allowing delete and insert."""

    # Notes here because GameDisplay instances used extensively to diagnose
    # problem.
    # Open a game with variations.  Analysis is affected too.
    # Navigate to a move where 'Down' cycles through available variations.
    # Use pointer to go somewhere else where 'Down' goes to next move.  The
    # pointer action can be over game score or board.
    # Use 'Up' or 'Down' to go to adjacent move.
    # This causes an exception and crashes ChessTab.
    # There is no crash if 'Up' is used to move away from a point where
    # cycling through variations occurs, or just to get out of cycling and
    # then use pointer.
    # Investigation suggests a rewrite of the management of keystroke and
    # pointer events is needed to fix the problem.

    # Allow for structure difference between GameDisplay and RepertoireDisplay
    # versions of delete_game_database.
    # Method comments suggest a problem exists which needs fixing.
    # Existence of this method prevents delete_game_database being used by
    # instances of superclasses of RepertoireDisplay, emulating the
    # behaviour before introduction of displaypgn module.
    @staticmethod
    def pgn_score_original_value(original_value):
        """Set game source as 'Editor' if error comment in original_value."""
        # currently attracts "AttributeError: 'ChessDBvalueGameTags' has no
        # attribute 'gamesource'.
        # original.value.set_game_source(self.sourceobject.value.gamesource)
        # original.value.set_game_source('Copy, possibly edited')
        if original_value.is_error_comment_present():
            original_value.set_game_source("Editor")

    def _create_primary_activity_popup(self):
        """Delegate then add close item entry and return popup menu."""
        popup = super()._create_primary_activity_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup

    def _create_select_move_popup(self):
        """Delegate then add close item entry and return popup menu."""
        popup = super()._create_select_move_popup()
        self._add_close_item_entry_to_popup(popup)
        return popup


class GameDisplayInsert(InsertPGN, _GameDisplay, GameEdit, DataNotify):
    """Display a chess game from a database allowing insert.

    GameEdit provides the widget and _GameDisplay the database interface.

    """


class GameDisplayEdit(EditPGN, GameDisplayInsert):
    """Display a chess game from a database allowing edit and insert."""

    # Allow for structure difference between GameDisplay and RepertoireDisplay
    # versions of delete_game_database.
    # Method comments suggest a problem exists which needs fixing.
    # Existence of this method prevents delete_game_database being used by
    # instances of superclasses of RepertoireDisplay, emulating the behaviour
    # before introduction of displaypgn module.
    @staticmethod
    def pgn_score_original_value(original_value):
        """Set game source as 'Editor' if error comment in original_value."""
        # currently attracts "AttributeError: 'ChessDBvalueGameTags' has
        # no attribute 'gamesource'.
        # original.value.set_game_source(self.sourceobject.value.gamesource)
        # original.value.set_game_source('Copy, possibly edited')
        if original_value.is_error_comment_present():
            original_value.set_game_source("Editor")

    # _set_properties_on_grids defined so update_game_database method can be
    # shared by repertoiredisplay.RepertoireDisplayEdit and
    # gamedisplay.GameDisplayEdit classes.
    # See class attributes pgn_score_name and pgn_score_source_name too.
    def _set_properties_on_grids(self, newkey):
        """Set properties of widgets for newkey on all grids."""
        self.ui.set_properties_on_all_game_grids(newkey)



# chessberkeleydbdu.py
# Copyright 2021 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using custom deferred update for Berkeley DB."""

import berkeleydb.db

from solentware_base import berkeleydbdu_database

from ..shared.dbdu import Dbdu
from ..shared.alldu import chess_du, Alldu


class ChessberkeleydbduError(Exception):
    """Exception class for chessberkeleydbdu module."""


def chess_database_du(dbpath, *args, estimated_number_of_games=0, **kwargs):
    """Open database, import games and close database."""
    chess_du(ChessDatabase(dbpath, allowcreate=True), *args, **kwargs)

    # There are no recoverable file full conditions for Berkeley DB (see DPT).
    return True


class ChessDatabase(Alldu, Dbdu, berkeleydbdu_database.Database):
    """Provide custom deferred update for a database of games of chess."""

    def __init__(self, DBfile, **kargs):
        """Delegate with ChessberkeleydbduError as exception class."""
        super().__init__(
            DBfile,
            ChessberkeleydbduError,
            (
                berkeleydb.db.DB_CREATE
                | berkeleydb.db.DB_RECOVER
                | berkeleydb.db.DB_INIT_MPOOL
                | berkeleydb.db.DB_INIT_LOCK
                | berkeleydb.db.DB_INIT_LOG
                | berkeleydb.db.DB_INIT_TXN
                | berkeleydb.db.DB_PRIVATE
            ),
            **kargs
        )



# gamelistgrid.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Grids for listing details of games on chess database."""

import tkinter.messagebox

from pgn_read.core.parser import PGN

from ..core.chessrecord import (
    ChessDBrecordRepertoireUpdate,
)
from .positionrow import ChessDBrowPosition
from .repertoirerow import ChessDBrowRepertoire
from .repertoiredisplay import RepertoireDisplay, RepertoireDisplayEdit
from .constants import (
    STATUS_SEVEN_TAG_ROSTER_EVENT,
    EMPTY_REPERTOIRE_GAME,
)
from ..core import export_repertoire
from .eventspec import EventSpec
from ..core.constants import REPERTOIRE_TAG_ORDER, UNKNOWN_RESULT
from .gamelistgrid import GameListGrid


class RepertoireGrid(GameListGrid):
    """Customized GameListGrid for list of repertoires on database."""

    def __init__(self, ui):
        """Extend with definition and bindings for games on database grid.

        ui - container for user interface widgets and methods.

        """
        super().__init__(ui.repertoires_pw, ui)
        self.make_header(ChessDBrowRepertoire.header_specification)
        self.__bind_on()
        self._set_popup_bindings(
            self.menupopup,
            (
                (
                    EventSpec.display_record_from_grid,
                    self._display_game_from_popup,
                ),
                (EventSpec.edit_record_from_grid, self._edit_game_from_popup),
            ),
        )
        self._add_cascade_menu_to_popup(
            "Export",
            self.menupopup,
            (
                (
                    EventSpec.pgn_export_format_no_comments,
                    self._export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self._export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self._export_text),
            ),
        )
        bindings = (
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (
                EventSpec.navigate_to_active_game,
                self._set_focus_gamepanel_item_command,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_active_repertoire,
                self._set_focus_repertoirepanel_item_command,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_partial,
                self._set_focus_partialpanel_item_command,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self._set_focus_selectionpanel_item_command,
            ),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
        )
        self._add_cascade_menu_to_popup("Navigation", self.menupopup, bindings)
        self._add_cascade_menu_to_popup(
            "Navigation", self.menupopupnorow, bindings
        )

    def bind_off(self):
        """Disable all bindings."""
        super().bind_off()
        self._set_event_bindings_frame(
            (
                (EventSpec.navigate_to_active_repertoire, ""),
                (EventSpec.navigate_to_repertoire_game_grid, ""),
                (EventSpec.navigate_to_partial_grid, ""),
                (EventSpec.navigate_to_active_partial, ""),
                (EventSpec.navigate_to_partial_game_grid, ""),
                (EventSpec.navigate_to_position_grid, ""),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, ""),
                (EventSpec.navigate_to_selection_rule_grid, ""),
                (EventSpec.navigate_to_active_selection_rule, ""),
                (EventSpec.display_record_from_grid, ""),
                (EventSpec.edit_record_from_grid, ""),
                (EventSpec.pgn_export_format_no_comments, ""),
                (EventSpec.pgn_export_format, ""),
                (EventSpec.pgn_import_format, ""),
                (EventSpec.text_internal_format, ""),
            )
        )

    def bind_on(self):
        """Enable all bindings."""
        super().bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self._set_event_bindings_frame(
            (
                (
                    EventSpec.navigate_to_active_repertoire,
                    self.set_focus_repertoirepanel_item,
                ),
                (
                    EventSpec.navigate_to_repertoire_game_grid,
                    self.set_focus_repertoire_game_grid,
                ),
                (
                    EventSpec.navigate_to_partial_grid,
                    self.set_focus_partial_grid,
                ),
                (
                    EventSpec.navigate_to_active_partial,
                    self.set_focus_partialpanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_game_grid,
                    self.set_focus_partial_game_grid,
                ),
                (
                    EventSpec.navigate_to_position_grid,
                    self.set_focus_position_grid,
                ),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
                (
                    EventSpec.navigate_to_selection_rule_grid,
                    self.set_focus_selection_rule_grid,
                ),
                (
                    EventSpec.navigate_to_active_selection_rule,
                    self.set_focus_selectionpanel_item,
                ),
                (EventSpec.display_record_from_grid, self._display_game),
                (EventSpec.edit_record_from_grid, self._edit_game),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self._export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self._export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self._export_text),
            )
        )

    def _display_game(self, event=None):
        """Display selected repertoire and cancel selection."""
        del event
        self._display_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def _display_game_from_popup(self, event=None):
        """Display repertoire selected by pointer."""
        del event
        self._display_selected_item(self.pointer_popup_selection)

    def _edit_game(self, event=None):
        """Display selected repertoire for editing and cancel selection."""
        del event
        self._edit_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def _edit_game_from_popup(self, event=None):
        """Display repertoire with editing allowed selected by pointer."""
        del event
        self._edit_selected_item(self.pointer_popup_selection)

    def on_game_change(self, instance):
        """Delegate to superclass if data source exists."""
        # may turn out to be just to catch datasource is None
        if self.get_data_source() is None:
            return
        super().on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                tags = self.objects[ss0].value.collected_game.pgn_tags
                self.ui.statusbar.set_status_text(
                    "  ".join([tags.get(k, "") for k in REPERTOIRE_TAG_ORDER])
                )
        else:
            self.ui.statusbar.set_status_text("")

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordRepertoireUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_delete_dialog(
            self.objects[key], oldobject, modal, title="Delete Repertoire"
        )

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordRepertoireUpdate(),
            ChessDBrecordRepertoireUpdate(),
            False,
            modal,
            title="Edit Repertoire",
        )

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordRepertoireUpdate(),
            ChessDBrecordRepertoireUpdate(),
            True,
            modal,
            title="Edit Repertoire",
        )

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        newobject = ChessDBrecordRepertoireUpdate()
        instance = self.datasource.new_row()
        instance.srvalue = repr(EMPTY_REPERTOIRE_GAME + UNKNOWN_RESULT)
        self.create_edit_dialog(
            instance, newobject, None, False, modal, title="New Repertoire"
        )

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordRepertoireUpdate()
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_show_dialog(
            self.objects[key], oldobject, modal, title="Show Repertoire"
        )

    def make_display_widget(self, sourceobject):
        """Return a RepertoireDisplay for sourceobject."""
        game = RepertoireDisplay(
            master=self.ui.view_repertoires_pw,
            ui=self.ui,
            items_manager=self.ui.repertoire_items,
            itemgrid=self.ui.repertoire_games,
            sourceobject=sourceobject,
        )
        game.set_position_analysis_data_source()
        game.collected_game = next(
            PGN(game_class=game.gameclass).read_games(
                sourceobject.get_srvalue()
            )
        )
        game.set_and_tag_item_text()
        return game

    def make_edit_widget(self, sourceobject):
        """Return a RepertoireDisplayEdit for sourceobject."""
        game = RepertoireDisplayEdit(
            master=self.ui.view_repertoires_pw,
            ui=self.ui,
            items_manager=self.ui.repertoire_items,
            itemgrid=self.ui.repertoire_games,
            sourceobject=sourceobject,
        )
        game.set_position_analysis_data_source()
        game.collected_game = next(
            PGN(game_class=game.gameclass).read_games(
                sourceobject.get_srvalue()
            )
        )
        game.set_and_tag_item_text(reset_undo=True)
        return game

    def _display_selected_item(self, key):
        """Display and return a RepertoireDisplay for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to GameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # GameDisplay is to be put.
        # Yes because GameDisplayEdit (see _edit_selected_item) includes
        # extra widgets. Want to say game.widget.destroy() eventually.
        # Read make_display_widget for GameDisplay and GameDisplayEdit.
        game = self.make_display_widget(selected)
        self.ui.add_repertoire_to_display(game)
        self.ui.repertoire_items.increment_object_count(key)
        self.ui.repertoire_items.set_itemmap(game, key)
        self.set_properties(key)
        return game

    def _edit_selected_item(self, key):
        """Display and return a RepertoireDisplayEdit for selected game."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and score be created here and
        # passed to GameDisplay. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # GameDisplayEdit is to be put.
        # Yes because GameDisplay (see _display_selected_item) includes
        # less widgets. Want to say game.widget.destroy() eventually.
        # Read make_edit_widget for GameDisplay and GameDisplayEdit.
        game = self.make_edit_widget(selected)
        self.ui.add_repertoire_to_display(game)
        self.ui.repertoire_items.increment_object_count(key)
        self.ui.repertoire_items.set_itemmap(game, key)
        self.set_properties(key)
        return game

    def _export_pgn(self, event=None):
        """Export selected repertoires in PGN export format."""
        self.ui.export_report(
            export_repertoire.export_selected_repertoires_pgn(
                self, self.ui.get_export_filename("Repertoires", pgn=True)
            ),
            "Repertoires",
        )

    def _export_pgn_no_comments(self, event=None):
        """Export selected repertoires in PGN export format no comments."""
        self.ui.export_report(
            export_repertoire.export_selected_repertoires_pgn_no_comments(
                self,
                self.ui.get_export_filename(
                    "Repertoires (no comments)", pgn=True
                ),
            ),
            "Repertoires (no comments)",
        )

    def _export_text(self, event=None):
        """Export selected repertoires as text."""
        self.ui.export_report(
            export_repertoire.export_selected_repertoires_text(
                self,
                self.ui.get_export_filename(
                    "Repertoires (internal format)", pgn=False
                ),
            ),
            "Repertoires (internal format)",
        )

    def export_pgn_import_format(self, event=None):
        """Export selected games in a PGN import format."""
        self.ui.export_report(
            export_repertoire.export_selected_repertoires_pgn_import_format(
                self,
                self.ui.get_export_filename(
                    "Repertoires (import format)", pgn=True
                ),
            ),
            "Repertoires (import format)",
        )

    def is_visible(self):
        """Return True if list of repertoire games is displayed."""
        return str(self.get_frame()) in self.ui.repertoires_pw.panes()

    def set_properties(self, key, dodefaultaction=True):
        """Return True if properties for game key set or False."""
        # Skip the immediate superclass method to not touch self.game_items
        # pylint bad-super-call message given.
        if super(GameListGrid, self).set_properties(
            key, dodefaultaction=False
        ):
            return True
        if self.ui.repertoire_items.object_display_count(key):
            self._set_background_on_display_row_under_pointer(key)
            return True
        if dodefaultaction:
            self._set_background_normal_row_under_pointer(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None."""
        # Skip the immediate superclass method to not touch self.game_items
        # pylint bad-super-call message given.
        row = super(GameListGrid, self).set_row(
            key, dodefaultaction=False, **kargs
        )
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if self.ui.repertoire_items.object_display_count(key):
            return self.objects[key].grid_row_on_display(**kargs)
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        return None

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_normal(
            self.ui.move_to_repertoire, self.ui.filter_repertoire
        )

    def set_selection(self, key):
        """Hack to fix edge case when inserting records using apsw or sqlite3.

        Workaround a KeyError exception when a record is inserted while a grid
        keyed by a secondary index with only one key value in the index is on
        display.

        """
        try:
            super().set_selection(key)
        except KeyError:
            tkinter.messagebox.showinfo(
                parent=self.parent,
                title="Insert Repertoire Workaround",
                message="".join(
                    (
                        "All records have same name on this display.\n\nThe ",
                        "new record has been inserted but you need to Hide, ",
                        "and then Show, the display to see the record in ",
                        "the list.",
                    )
                ),
            )


class RepertoirePositionGames(GameListGrid):
    """Customized GameListGrid for list of games matching repertoire position.

    The grid is populated by a FullPositionDS instance from the
    dpt.fullpositionds or basecore.fullpositionds modules.
    """

    def __init__(self, ui):
        """Extend with position grid definition and bindings.

        ui - container for user interface widgets and methods.

        """
        super().__init__(ui.position_repertoires_pw, ui)
        self.make_header(ChessDBrowPosition.header_specification)
        self.__bind_on()
        self._set_popup_bindings(
            self.menupopup,
            (
                (
                    EventSpec.display_record_from_grid,
                    self._display_game_from_popup,
                ),
                (EventSpec.edit_record_from_grid, self._edit_game_from_popup),
            ),
        )
        self._add_cascade_menu_to_popup(
            "Export",
            self.menupopup,
            (
                (
                    EventSpec.pgn_reduced_export_format,
                    self._export_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self._export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self._export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self._export_text),
            ),
        )
        bindings = (
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (
                EventSpec.navigate_to_active_game,
                self._set_focus_gamepanel_item_command,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self._set_focus_repertoirepanel_item_command,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_partial,
                self._set_focus_partialpanel_item_command,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (
                EventSpec.navigate_to_selection_rule_grid,
                self.set_focus_selection_rule_grid,
            ),
            (
                EventSpec.navigate_to_active_selection_rule,
                self._set_focus_selectionpanel_item_command,
            ),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
        )
        self._add_cascade_menu_to_popup("Navigation", self.menupopup, bindings)
        self._add_cascade_menu_to_popup(
            "Navigation", self.menupopupnorow, bindings
        )

    def bind_off(self):
        """Disable all bindings."""
        super().bind_off()
        self._set_event_bindings_frame(
            (
                (EventSpec.navigate_to_repertoire_grid, ""),
                (EventSpec.navigate_to_active_repertoire, ""),
                (EventSpec.navigate_to_partial_grid, ""),
                (EventSpec.navigate_to_active_partial, ""),
                (EventSpec.navigate_to_partial_game_grid, ""),
                (EventSpec.navigate_to_position_grid, ""),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, ""),
                (EventSpec.navigate_to_selection_rule_grid, ""),
                (EventSpec.navigate_to_active_selection_rule, ""),
                (EventSpec.display_record_from_grid, ""),
                (EventSpec.edit_record_from_grid, ""),
                (EventSpec.pgn_reduced_export_format, ""),
                (EventSpec.pgn_export_format_no_comments, ""),
                (EventSpec.pgn_export_format, ""),
            )
        )

    def bind_on(self):
        """Enable all bindings."""
        super().bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self._set_event_bindings_frame(
            (
                (
                    EventSpec.navigate_to_repertoire_grid,
                    self.set_focus_repertoire_grid,
                ),
                (
                    EventSpec.navigate_to_active_repertoire,
                    self.set_focus_repertoirepanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_grid,
                    self.set_focus_partial_grid,
                ),
                (
                    EventSpec.navigate_to_active_partial,
                    self.set_focus_partialpanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_game_grid,
                    self.set_focus_partial_game_grid,
                ),
                (
                    EventSpec.navigate_to_position_grid,
                    self.set_focus_position_grid,
                ),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
                (
                    EventSpec.navigate_to_selection_rule_grid,
                    self.set_focus_selection_rule_grid,
                ),
                (
                    EventSpec.navigate_to_active_selection_rule,
                    self.set_focus_selectionpanel_item,
                ),
                (EventSpec.display_record_from_grid, self._display_game),
                (EventSpec.edit_record_from_grid, self._edit_game),
                (
                    EventSpec.pgn_reduced_export_format,
                    self._export_pgn_reduced_export_format,
                ),
                (
                    EventSpec.pgn_export_format_no_comments,
                    self._export_pgn_no_comments,
                ),
                (EventSpec.pgn_export_format, self._export_pgn),
                (EventSpec.pgn_import_format, self.export_pgn_import_format),
                (EventSpec.text_internal_format, self._export_text),
            )
        )

    def _display_game(self, event=None):
        """Display selected game and cancel selection."""
        del event
        self.set_move_highlight(
            self._display_selected_item(self.get_visible_selected_key())
        )
        self.cancel_selection()

    def _display_game_from_popup(self, event=None):
        """Display game selected by pointer."""
        del event
        self.set_move_highlight(
            self._display_selected_item(self.pointer_popup_selection)
        )

    def _edit_game(self, event=None):
        """Display selected game with editing allowed and cancel selection."""
        del event
        self.set_move_highlight(
            self._edit_selected_item(self.get_visible_selected_key())
        )
        self.cancel_selection()

    def _edit_game_from_popup(self, event=None):
        """Display game with editing allowed selected by pointer."""
        del event
        self.set_move_highlight(
            self._edit_selected_item(self.pointer_popup_selection)
        )

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for game key or None.

        Add arguments to **kargs for grid_row method in PositionRow class.

        """
        kargs.update(
            position=self.datasource.fullposition,
            context=self.ui.get_active_repertoire_move(),
        )
        return super().set_row(key, dodefaultaction=dodefaultaction, **kargs)

    def on_game_change(self, instance):
        """Delegate to superclass if data source exists."""
        # datasource refers to a set derived from file and may need
        # to be recreated
        if self.get_data_source() is None:
            return
        super().on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display main PGN Tags."""
        if self.selection:
            ss0 = self.selection[0]
            if ss0 in self.objects:
                tags = self.objects[ss0].score.collected_game.pgn_tags
                self.ui.statusbar.set_status_text(
                    "  ".join(
                        [
                            tags.get(k, "")
                            for k in STATUS_SEVEN_TAG_ROSTER_EVENT
                        ]
                    )
                )
        else:
            self.ui.statusbar.set_status_text("")

    def is_visible(self):
        """Return True if list of matching repertoire games is displayed."""
        return str(self.get_frame()) in self.ui.position_repertoires_pw.panes()

    def is_payload_available(self):
        """Return True if connected to database and games displayed."""
        if not super().is_payload_available():
            return False
        return self.ui.repertoire_items.is_visible()

    def make_display_widget(self, sourceobject):
        """Return a GameDisplay for sourceobject."""
        game = super().make_display_widget(sourceobject)
        game.set_and_tag_item_text()
        return game

    def make_edit_widget(self, sourceobject):
        """Return a GameDisplayEdit for sourceobject."""
        game = super().make_edit_widget(sourceobject)
        game.set_and_tag_item_text(reset_undo=True)
        return game

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_disabled()



# eventspec.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Map ChessTab event names to tk(inter) event detail values."""

# Some cases need an object with a widget attribute, but the source of the
# event does not provide an event object, let alone the relevant widget.
# Fortunately the required widget is always known in these cases.
import collections

DummyEvent = collections.namedtuple("DummyEvent", "widget")
del collections

# Mapped to widget focus setters and initialisers in all panels.
# Menu item label is same in all cases.
GAMES_PARTIAL_POSITION = "Games Partial Position"
ACTIVE_PARTIAL_POSITION = "Active Partial Position"
PARTIAL_POSITION_LIST = "Partial Position List"
GAMES_REPERTOIRE = "Games Repertoire"
ACTIVE_REPERTOIRE = "Active Repertoire"
REPERTOIRE_GAME_LIST = "Repertoire Game List"
GAMES_DATABASE = "Games Database"
POSITION_GAME_LIST = "Position Game List"
ACTIVE_GAME = "Active Game"
SELECTION_RULE_LIST = "Selection Rule List"
ACTIVE_SELECTION_RULE = "Active Selection Rule"

# Mapped to widget focus setters and initialisers in relevant panels.
# Menu item label is same in all cases.
PREVIOUS_REPERTOIRE = "Previous Repertoire"
NEXT_REPERTOIRE = "Next Repertoire"
REPERTOIRE = "Repertoire"
PREVIOUS_GAME = "Previous Game"
NEXT_GAME = "Next Game"
PREVIOUS_PARTIAL_POSITION = "Previous Partial Position"
NEXT_PARTIAL_POSITION = "Next Partial Position"
PREVIOUS_SELECTION_RULE = "Previous Selection Rule"
NEXT_SELECTION_RULE = "Next Selection Rule"


class EventSpec:
    """Event detail values for ChessTab keyboard and pointer actions."""

    # Navigate to another specific widget event specifications.  (Not next
    # widget in tab order or similar.)
    # When this class was introduced it was convenient name the events like
    # 'position_grid_to_partial_grid' and 'partial_game_grid_to_partial_grid',
    # though both were "('<KeyPress-F2>', PARTIAL_POSITION_LIST, 'F2')", to
    # avoid making mistakes.  Opportunity taken, when adding export events,
    # to replace these with 'navigate_to_partial_grid' names.
    navigate_to_active_game = ("<Alt-KeyPress-F9>", ACTIVE_GAME, "Alt F9")
    navigate_to_active_partial = (
        "<Alt-KeyPress-F2>",
        ACTIVE_PARTIAL_POSITION,
        "Alt F2",
    )
    navigate_to_active_repertoire = (
        "<Alt-KeyPress-F6>",
        ACTIVE_REPERTOIRE,
        "Alt F6",
    )
    navigate_to_active_selection_rule = (
        "<Alt-KeyPress-F3>",
        ACTIVE_SELECTION_RULE,
        "Alt F3",
    )
    navigate_to_game_grid = ("<KeyPress-F9>", GAMES_DATABASE, "F9")
    navigate_to_partial_game_grid = (
        "<Shift-KeyPress-F2>",
        GAMES_PARTIAL_POSITION,
        "Shift F2",
    )
    navigate_to_partial_grid = ("<KeyPress-F2>", PARTIAL_POSITION_LIST, "F2")
    navigate_to_position_grid = (
        "<Shift-KeyPress-F9>",
        POSITION_GAME_LIST,
        "Shift F9",
    )
    navigate_to_repertoire_game_grid = (
        "<Shift-KeyPress-F6>",
        GAMES_REPERTOIRE,
        "Shift F6",
    )
    navigate_to_repertoire_grid = ("<KeyPress-F6>", REPERTOIRE_GAME_LIST, "F6")
    navigate_to_selection_rule_grid = (
        "<KeyPress-F3>",
        SELECTION_RULE_LIST,
        "F3",
    )

    # All the grids have Display and Display allow edit options, with the same
    # prompt and keystrokes no matter the type of data.  The grids are
    # PartialPositionGames, GamePositionGames, TagRosterGrid, RepertoireGrid,
    # RepertoirePositionGames, PartialGrid, and SelectionGrid.
    display_record_from_grid = ("<KeyPress-F11>", "Display", "F11")
    edit_record_from_grid = (
        "<Control-KeyPress-F11>",
        "Display allow edit",
        "Ctrl F11",
    )

    # PartialGrid
    export_from_partial_grid = (
        "<Control-Alt-KeyPress-Home>",
        "Export",
        "Ctrl Alt Home",
    )

    # All the display, and edit, widgets have Insert, Delete, Update, and Make
    # Active options.  These event definitions replace those named for the
    # widget using them.
    # The Dismiss option is a plausible renaming of the Close options named
    # after the data in the widget.
    # Some have a List Games option.
    display_insert = ("<KeyPress-Insert>", "Insert", "Insert")
    display_delete = ("<KeyPress-Delete>", "Delete", "Delete")
    display_update = ("<Alt-KeyPress-Insert>", "Edit", "Alt Insert")
    display_dismiss = ("<Alt-KeyPress-F12>", "Close Item", "Alt F12")
    display_make_active = ("", "Make Active", "Left Click")
    display_dismiss_inactive = ("", "Close Item", "")
    display_list = ("<Control-KeyPress-Return>", "List Games", "Ctrl Enter")

    # Score widget, common to Game and Repertoire widgets.
    # Default lambda in Score.__init__ ignored.
    score_show_previous_in_variation = (
        "<Shift-KeyPress-Up>",
        "Previous Move Variation",
        "Shift Up",
    )
    score_show_previous_in_line = ("<KeyPress-Up>", "Previous Move", "Up")
    score_show_next_in_line = (
        "<Shift-KeyPress-Down>",
        "Next Move",
        "Shift Down",
    )
    score_show_next_in_variation = (
        "<KeyPress-Down>",
        "Next Move Select Variation",
        "Down",
    )
    score_show_first_in_line = (
        "<KeyPress-Prior>",
        "Start of Variation",
        "PageUp",
    )
    score_show_last_in_line = ("<KeyPress-Next>", "End of Variation", "PageDn")
    score_show_first_in_game = (
        "<Shift-KeyPress-Prior>",
        "Start of Game",
        "Shift PageUp",
    )
    score_show_last_in_game = (
        "<Shift-KeyPress-Next>",
        "End of Game",
        "Shift PageDn",
    )
    score_show_selected_variation = (
        "<Shift-KeyPress-Down>",
        "Enter Variation",
        "Shift Down",
    )
    score_cycle_selection_to_next_variation = (
        "<KeyPress-Down>",
        "Next Variation",
        "Down",
    )
    score_cancel_selection_of_variation = (
        "<KeyPress-Up>",
        "Cancel Variation",
        "Up",
    )
    score_disable_keypress = (
        "<KeyPress>",
        "",
        "",
    )  # No menu entry for this initialisation event.
    score_enable_F10_menubar = ("<KeyPress-F10>", "", "F10")
    score_enable_F10_popupmenu_at_top_left = (
        "<Shift-KeyPress-F10>",
        "Popup Menu at Top Left",
        "Shift F10",
    )
    score_enable_F10_popupmenu_at_pointer = (
        "<Control-KeyPress-F10>",
        "Popup Menu at Pointer",
        "Ctrl F10",
    )

    # Game widget analyse events.
    # Inherited by Repertoire, GameEdit, and RepertoireEdit, widgets.
    analyse_game = ("<Alt-KeyPress-a>", "Analyse", "Alt a")

    # GameEdit widget editing events.
    # Inherited by RepertoireEdit widget.
    gameedit_insert_rav = (
        "<KeyPress>",
        "Insert RAV: move move (<char>)",
        "<char>",
    )
    gameedit_insert_rav_after_rav_end = (
        "<KeyPress>",
        "Insert RAV: (...)(<char>)",
        "<char>",
    )
    gameedit_insert_rav_after_rav_start_move_or_rav = (
        "<Shift-KeyPress>",
        "Insert RAV: ( move (<char>)",
        "Shift <char>",
    )
    gameedit_insert_rav_after_rav_start = (
        "<Alt-KeyPress>",
        "Insert RAV: ((<char>) move",
        "Alt <char>",
    )
    gameedit_insert_castle_queenside = (
        "<Control-KeyPress-o>",
        "O-O-O",
        "Ctrl o",
    )
    gameedit_insert_comment = (
        "<Control-KeyPress-braceleft>",
        "Insert Comment",
        "Ctrl {",
    )
    gameedit_insert_reserved = (
        "<Control-KeyPress-less>",
        "Insert Reserved",
        "Ctrl <",
    )
    gameedit_insert_comment_to_eol = (
        "<Control-KeyPress-semicolon>",
        "Insert Comment to EOL",
        "Ctrl ;",
    )
    gameedit_insert_escape_to_eol = (
        "<Control-KeyPress-percent>",
        "Insert Escape to EOL",
        "Ctrl %",
    )
    gameedit_insert_glyph = (
        "<Control-KeyPress-dollar>",
        "Insert Glyph",
        "Ctrl $",
    )
    gameedit_insert_pgn_tag = (
        "<Control-KeyPress-bracketleft>",
        "Insert PGN Tag",
        "Ctrl [",
    )
    gameedit_insert_pgn_seven_tag_roster = (
        "<Control-KeyPress-bracketright>",
        "Insert PGN Seven Tag Roster",
        "Ctrl ]",
    )
    gameedit_insert_white_win = ("<Control-KeyPress-plus>", "1-0", "Ctrl +")
    gameedit_insert_draw = ("<Control-KeyPress-equal>", "1/2-1/2", "Ctrl =")
    gameedit_insert_black_win = ("<Control-KeyPress-minus>", "0-1", "Ctrl -")
    gameedit_insert_other_result = (
        "<Control-KeyPress-asterisk>",
        "Other Result",
        "Ctrl *",
    )
    gameedit_show_previous_token = (
        "<Shift-KeyPress-Left>",
        "Previous Item",
        "Shift Left",
    )
    gameedit_show_next_token = (
        "<Shift-KeyPress-Right>",
        "Next Item",
        "Shift Right",
    )
    gameedit_show_first_token = (
        "<Shift-KeyPress-Prior>",
        "First Item",
        "Shift PageUp",
    )
    gameedit_show_last_token = (
        "<Shift-KeyPress-Next>",
        "Last Item",
        "Shift PageDn",
    )
    gameedit_show_first_comment = (
        "<Control-KeyPress-Prior>",
        "First Comment",
        "Ctrl PageUp",
    )
    gameedit_show_last_comment = (
        "<Control-KeyPress-Next>",
        "Last Comment",
        "Ctrl PageDn",
    )
    gameedit_show_previous_comment = (
        "<Control-KeyPress-Up>",
        "Previous Comment",
        "Ctrl Up",
    )
    gameedit_show_next_comment = (
        "<Control-KeyPress-Down>",
        "Next Comment",
        "Ctrl Down",
    )
    gameedit_show_previous_rav_start = (
        "<Alt-KeyPress-Prior>",
        "Previous RAV Start",
        "Alt PageUp",
    )
    gameedit_show_next_rav_start = (
        "<Alt-KeyPress-Next>",
        "Next RAV Start",
        "Alt PageDn",
    )
    gameedit_to_previous_pgn_tag = (
        "<Control-KeyPress-Left>",
        "Previous PGN Tag",
        "Ctrl Left",
    )
    gameedit_to_next_pgn_tag = (
        "<Control-KeyPress-Right>",
        "Next PGN Tag",
        "Ctrl Right",
    )
    gameedit_delete_empty_pgn_tag = (
        "<Control-KeyPress-Delete>",
        "Delete empty PGN Tag",
        "Ctrl Delete",
    )
    gameedit_bind_and_show_previous_in_line = (
        "<KeyPress-Up>",
        "Previous Move",
        "Up",
    )
    gameedit_bind_and_show_previous_in_variation = (
        "<Shift-KeyPress-Up>",
        "Previous Move Variation",
        "Shift Up",
    )
    gameedit_bind_and_show_next_in_line = (
        "<Shift-KeyPress-Down>",
        "Next Move",
        "Shift Down",
    )
    gameedit_bind_and_show_next_in_variation = (
        "<KeyPress-Down>",
        "Next Move Select Variation",
        "Down",
    )
    gameedit_bind_and_show_first_in_line = (
        "<KeyPress-Prior>",
        "Start of Variation",
        "PageUp",
    )
    gameedit_bind_and_show_last_in_line = (
        "<KeyPress-Next>",
        "End of Variation",
        "PageDn",
    )
    gameedit_bind_and_show_first_in_game = (
        "<Shift-KeyPress-Prior>",
        "Start of Game",
        "Shift PageUp",
    )
    gameedit_bind_and_show_last_in_game = (
        "<Shift-KeyPress-Next>",
        "End of Game",
        "Shift PageDn",
    )
    gameedit_bind_and_to_previous_pgn_tag = (
        "<Control-KeyPress-Left>",
        "Previous PGN Tag",
        "Ctrl Left",
    )
    gameedit_bind_and_to_next_pgn_tag = (
        "<Control-KeyPress-Right>",
        "Next PGN Tag",
        "Ctrl Right",
    )
    gameedit_delete_token_char_left = (
        "<Shift-KeyPress-BackSpace>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_delete_token_char_right = (
        "<Shift-KeyPress-Delete>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_delete_char_left = (
        "<KeyPress-BackSpace>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_delete_char_right = (
        "<KeyPress-Delete>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_set_insert_previous_line_in_token = (
        "<Alt-KeyPress-Up>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_set_insert_previous_char_in_token = (
        "<KeyPress-Left>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_set_insert_next_char_in_token = (
        "<KeyPress-Right>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_set_insert_next_line_in_token = (
        "<Alt-KeyPress-Down>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_set_insert_first_char_in_token = (
        "<KeyPress-Home>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_set_insert_last_char_in_token = (
        "<KeyPress-End>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_add_char_to_token = (
        "<KeyPress>",
        "",
        "",
    )  # A menu entry would say type text to insert.
    gameedit_non_move_show_previous_in_variation = (
        "<Shift-KeyPress-Up>",
        "Previous Move Variation",
        "Shift Up",
    )
    gameedit_non_move_show_previous_in_line = (
        "<KeyPress-Up>",
        "Previous Move",
        "Up",
    )
    gameedit_non_move_show_next_in_line = (
        "<Shift-KeyPress-Down>",
        "Next Move",
        "Shift Down",
    )
    gameedit_non_move_show_next_in_variation = (
        "<KeyPress-Down>",
        "Next Move Select Variation",
        "Down",
    )
    gameedit_non_move_show_first_in_line = (
        "<KeyPress-Prior>",
        "Start of Variation",
        "PageUp",
    )
    gameedit_non_move_show_last_in_line = (
        "<KeyPress-Next>",
        "End of Variation",
        "PageDn",
    )
    gameedit_non_move_show_first_in_game = (
        "<Shift-KeyPress-Prior>",
        "Start of Game",
        "Shift PageUp",
    )
    gameedit_non_move_show_last_in_game = (
        "<Shift-KeyPress-Next>",
        "End of Game",
        "Shift PageDn",
    )

    # Why specific shift versions?
    gameedit_delete_move_char_left_shift = (
        "<Shift-KeyPress-BackSpace>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_delete_move_char_right_shift = (
        "<Shift-KeyPress-Delete>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_delete_move_char_left = (
        "<KeyPress-BackSpace>",
        "",
        "",
    )  # A menu entry would make sense.
    gameedit_delete_move_char_right = (
        "<KeyPress-Delete>",
        "",
        "",
    )  # A menu entry would make sense.

    gameedit_add_move_char_to_token = (
        "<KeyPress>",
        "",
        "",
    )  # A menu entry would say type moves to insert.
    gameedit_insert_move = (
        "<KeyPress>",
        "",
        "",
    )  # A menu entry would say type moves to insert.
    gameedit_edit_move = (
        "<KeyPress-BackSpace>",
        "",
        "",
    )  # A menu entry would make sense.

    # GameDisplay
    gamedisplay_to_previous_game = ("<KeyPress-F7>", PREVIOUS_GAME, "F7")
    gamedisplay_to_next_game = ("<KeyPress-F8>", NEXT_GAME, "F8")

    # Switch between scoresheet and analysis in game or repertoire.
    scoresheet_to_analysis = ("<Alt-KeyPress-F8>", "Analysis", "Alt F8")
    analysis_to_scoresheet = ("<Alt-KeyPress-F8>", "Scoresheet", "Alt F8")

    # PartialDisplay
    partialdisplay_to_previous_partial = (
        "<KeyPress-F7>",
        PREVIOUS_PARTIAL_POSITION,
        "F7",
    )
    partialdisplay_to_next_partial = (
        "<KeyPress-F8>",
        NEXT_PARTIAL_POSITION,
        "F8",
    )
    export_from_partialdisplay = (
        "<Control-Alt-KeyPress-Home>",
        "Export",
        "Ctrl Alt Home",
    )

    # PartialEdit
    partialedit_insert_char_or_token = (
        "<KeyPress>",
        "",
        "",
    )  # No menu entry because character required.
    partialedit_delete_char_left = (
        "<KeyPress-BackSpace>",
        "",
        "",
    )  # A menu entry would make sense.
    partialedit_delete_char_right = (
        "<KeyPress-Delete>",
        "",
        "",
    )  # A menu entry would make sense.
    partialedit_show_previous_token = ("<KeyPress-Up>", "Previous Piece", "Up")
    partialedit_show_next_token = ("<KeyPress-Down>", "Next Piece", "Down")
    partialedit_show_first_token = (
        "<KeyPress-Prior>",
        "First Piece",
        "PageUp",
    )
    partialedit_show_last_token = ("<KeyPress-Next>", "Last Piece", "PageDn")
    partialedit_insert_partial_name_left = (
        "<Control-KeyPress-bracketleft>",
        "Name",
        "Ctrl [",
    )
    partialedit_insert_partial_name_right = (
        "<Control-KeyPress-bracketright>",
        "Name",
        "Ctrl ]",
    )
    partialedit_set_insert_previous_char_in_token = (
        "<KeyPress-Left>",
        "",
        "",
    )  # A menu entry would make sense.
    partialedit_set_insert_next_char_in_token = (
        "<KeyPress-Right>",
        "",
        "",
    )  # A menu entry would make sense.
    partialedit_set_insert_first_char_in_token = (
        "<KeyPress-Home>",
        "",
        "",
    )  # A menu entry would make sense.
    partialedit_set_insert_last_char_in_token = (
        "<KeyPress-End>",
        "",
        "",
    )  # A menu entry would make sense.

    # RepertoireDisplay
    repertoiredisplay_to_previous_repertoire = (
        "<KeyPress-F7>",
        PREVIOUS_REPERTOIRE,
        "F7",
    )
    repertoiredisplay_to_next_repertoire = (
        "<KeyPress-F8>",
        NEXT_REPERTOIRE,
        "F8",
    )

    # PartialScore
    partialscore_disable_keypress = (
        "<KeyPress>",
        "",
        "",
    )  # No menu entry because character required.

    # Partial widget partial position export events.
    # Inherited by PartialEdit widget.
    text_from_partial = (
        "<Control-Alt-KeyPress-Home>",
        "Text",
        "Ctrl Alt Home",
    )

    # Keyboard traversal events (Control-F8 Control-F7): all relevant widgets.
    # tab_traverse_backward bindings behave differently to expectation given
    # tab_traverse_forward behaviour at Python3.3 with Tk8.6, although I do not
    # know what happens with Tk8.5 an so forth.  Similarely <Shift-Tab>.
    tab_traverse_forward = ("<Control-KeyPress-F8>", "Focus Next", "Ctrl F8")
    tab_traverse_backward = (
        "<Control-KeyPress-F7>",
        "Focus Previous",
        "Ctrl F7",
    )
    tab_traverse_round = ("<Alt-KeyPress-F8>", "Focus Round", "Alt F8")

    # SelectionDisplay
    selectiondisplay_to_previous_selection = (
        "<KeyPress-F7>",
        PREVIOUS_SELECTION_RULE,
        "F7",
    )
    selectiondisplay_to_next_selection = (
        "<KeyPress-F8>",
        NEXT_SELECTION_RULE,
        "F8",
    )

    # SelectionText
    selectiontext_disable_keypress = (
        "<KeyPress>",
        "",
        "",
    )  # Keyboard actions do nothing by default.

    # EngineGrid
    engine_grid_run = ("<Alt-KeyPress-r>", "Run Engine", "Alt r")

    # EngineText
    enginetext_disable_keypress = (
        "<KeyPress>",
        "",
        "",
    )  # Keyboard actions do nothing by default.

    # DatabaseEngineDisplay
    databaseenginedisplay_run = ("<Alt-KeyPress-r>", "Run Engine", "Alt r")

    # DatabaseEngineEdit
    databaseengineedit_browse = ("<Alt-KeyPress-b>", "Browse Engines", "Alt b")

    # Export PGN options.
    # These are put in the 'Export' submenu available on all widgets based on
    # subclasses of Score or GameListGrid; and in the main Database menu.
    # Attempt with <Alt-Shift-KeyPress-Home> for reduced export format failed.
    # Shift-Home is already taken by the bindings defined for DataGrids in the
    # solentware_misc.gui.datagrid module: hence Shift-Escape below.
    pgn_reduced_export_format = (
        "<Alt-KeyPress-End>",
        "PGN reduced",
        "Alt End",
        4,
    )
    pgn_export_format_no_comments_no_ravs = (
        "<Control-KeyPress-Home>",
        "PGN all tags moves played",
        "Ctrl Home",
        4,
    )
    pgn_export_format_no_comments = (
        "<Control-Shift-KeyPress-Home>",
        "PGN no comments",
        "Ctrl Shift Home",
        4,
    )
    pgn_export_format = (
        "<Control-Alt-KeyPress-Home>",
        "PGN",
        "Ctrl Alt Home",
        0,
    )
    pgn_import_format = (
        "<Shift-KeyPress-Escape>",
        "PGN import",
        "Shift Escape",
        4,
    )
    text_internal_format = ("<Alt-KeyPress-Home>", "Text", "Alt Home", 0)

    # ButtonPress event definitions.
    control_buttonpress_1 = ("<Control-ButtonPress-1>", "", "")
    control_buttonpress_3 = ("<Control-ButtonPress-3>", "", "")
    shift_buttonpress_1 = ("<Shift-ButtonPress-1>", "", "")
    shift_buttonpress_3 = ("<Shift-ButtonPress-3>", "", "")
    alt_buttonpress_1 = ("<Alt-ButtonPress-1>", "", "")
    alt_buttonpress_3 = ("<Alt-ButtonPress-3>", "", "")
    buttonpress_1 = ("<ButtonPress-1>", "", "")
    buttonpress_3 = ("<ButtonPress-3>", "", "")

    # Menubar event definitions.
    # F10 in any window invokes the first menu (probably Database).
    # Alt-x invokes the corresponding menu in the menubar, where x is the
    # underlined character.
    menu_database_open = ("", "Open", "", 0)
    menu_database_new = ("", "New", "", 0)
    menu_database_close = ("", "Close", "", 0)
    menu_database_import = ("", "Import", "", 0)
    menu_database_export = ("", "Export", "", 0)
    menu_database_delete = ("", "Delete", "", 0)
    menu_database_quit = ("", "Quit", "", 0)
    menu_database_export_all_text = ("", "All (text)", "", 0)

    # Games, Repertoires, and Positions, are used in the Import and Export
    # submenus of the Database menu.
    menu_database_games = ("", "Games", "", 0)
    menu_database_repertoires = ("", "Repertoires", "", 0)
    menu_database_positions = ("", "Positions", "", 0)

    # Show and hide are used in the Select, Position, and Repertoire menus.
    menu_show = ("", "Show", "", 0)
    menu_hide = ("", "Hide", "", 0)

    menu_select_rule = ("", "Rule", "", 0)
    menu_select_game = ("", "Game", "", 0)
    menu_select_index = ("", "Index", "", 0)
    menu_select_error = ("", "Error", "", 0)

    # Submenu for menu_select_index.
    menu_select_index_black = ("", "Black", "", 0)
    menu_select_index_white = ("", "White", "", 0)
    menu_select_index_event = ("", "Event", "", 0)
    menu_select_index_date = ("", "Date", "", 0)
    menu_select_index_result = ("", "Result", "", 0)
    menu_select_index_site = ("", "Site", "", 0)
    menu_select_index_round = ("", "Round", "", 4)

    menu_game_new_game = ("", "New Game", "", 0)
    menu_position_partial = ("", "Partial", "", 0)
    menu_repertoire_opening = ("", "Opening", "", 0)
    menu_tools_board_style = ("", "Board Style", "", 6)
    menu_tools_board_fonts = ("", "Board Fonts", "", 6)
    menu_tools_board_colours = ("", "Board Colours", "", 6)
    menu_tools_hide_game_analysis = ("", "Hide Game Analysis", "", 0)
    menu_tools_show_game_analysis = ("", "Show Game Analysis", "", 5)
    menu_tools_hide_game_scrollbars = ("", "Hide Game Scrollbars", "", 1)
    menu_tools_show_game_scrollbars = ("", "Show Game Scrollbars", "", 2)
    menu_tools_toggle_game_move_numbers = (
        "",
        "Toggle Game Move Numbers",
        "",
        12,
    )
    menu_tools_toggle__analysis_fen = ("", "Toggle Analysis Fen", "", 7)
    menu_tools_toggle_single_view = ("", "Toggle Single View", "", 14)
    menu_engines_position_queues = ("", "Position Queues", "", 0)
    menu_engines_show_engines = ("", "Show Engines", "", 5)
    menu_engines_start_engine = ("", "Start Engine", "", 0)
    menu_engines_quit_all_engines = ("", "Quit All Engines", "", 0)
    menu_commands_multipv = ("", "MultiPV", "", 6)
    menu_commands_depth = ("", "Depth", "", 0)
    menu_commands_hash = ("", "Hash", "", 0)
    menu_help_guide = ("", "Guide", "", 0)
    menu_help_selection_rules = ("", "Selection rules", "", 0)
    menu_help_file_size = ("", "File size", "", 0)
    menu_help_notes = ("", "Notes", "", 0)
    menu_help_about = ("", "About", "", 0)



# querygrid.py
# Copyright 2015 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Grids for listing details of selection rules on chess database."""

import tkinter.messagebox

from solentware_grid.datagrid import DataGrid

from ..core.chessrecord import ChessDBrecordQuery
from .querydisplay import QueryDisplay, QueryDisplayEdit
from .queryrow import ChessDBrowQuery
from ..core import export_selection_rule
from .eventspec import EventSpec
from .display import Display
from ..shared.cql_gamelist_query import CQLGameListQuery
from ..shared.allgrid import AllGrid


class QueryListGrid(AllGrid, CQLGameListQuery, DataGrid, Display):
    """A DataGrid for lists of game selection rules.

    Subclasses provide navigation and extra methods appropriate to list use.

    """

    def __init__(self, parent, ui):
        """Extend with link to user interface object.

        parent - see superclass
        ui - container for user interface widgets and methods.

        """
        super().__init__(parent=parent)
        self.ui = ui
        self._configure_frame_and_initial_event_bindings()

    def _display_selected_item_kind(self, key, selected):
        # Should the Frame containing board and position be created here and
        # passed to QueryDisplay. (Needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # QueryDisplay is to be put.
        selection = self.make_display_widget(selected)
        self.ui.add_selection_rule_to_display(selection)
        self.ui.selection_items.increment_object_count(key)
        self.ui.selection_items.set_itemmap(selection, key)
        self.set_properties(key)
        return selection

    def make_display_widget(self, sourceobject):
        """Return a QueryDisplay for sourceobject."""
        selection = QueryDisplay(
            master=self.ui.view_selection_rules_pw,
            ui=self.ui,
            items_manager=self.ui.selection_items,
            itemgrid=self.ui.base_games,
            sourceobject=sourceobject,
        )
        selection.query_statement.set_database(
            self.ui.base_games.datasource.dbhome
        )
        selection.query_statement.dbset = self.ui.base_games.datasource.dbset
        selection.query_statement.process_query_statement(
            sourceobject.get_srvalue()
        )
        return selection

    def _edit_selected_item(self, key):
        """Create a QueryDisplayEdit for game selection rule."""
        selected = self.get_visible_record(key)
        if selected is None:
            return None
        # Should the Frame containing board and position be created here and
        # passed to QueryDisplayEdit. (Which needs 'import Tkinter' above.)
        # Rather than passing the container where the Frame created by
        # QueryDisplayEdit is to be put.
        selection = self.make_edit_widget(selected)
        self.ui.add_selection_rule_to_display(selection)
        self.ui.selection_items.increment_object_count(key)
        self.ui.selection_items.set_itemmap(selection, key)
        self.set_properties(key)
        return selection

    def make_edit_widget(self, sourceobject):
        """Return a QueryDisplayEdit for sourceobject."""
        selection = QueryDisplayEdit(
            master=self.ui.view_selection_rules_pw,
            ui=self.ui,
            items_manager=self.ui.selection_items,
            itemgrid=self.ui.base_games,
            sourceobject=sourceobject,
        )
        selection.query_statement.set_database(
            self.ui.base_games.datasource.dbhome
        )
        selection.query_statement.dbset = self.ui.base_games.datasource.dbset
        selection.query_statement.process_query_statement(
            sourceobject.get_srvalue()
        )
        return selection

    def set_properties(self, key, dodefaultaction=True):
        """Return True if properties for selection rule key set or False."""
        if super().set_properties(key, dodefaultaction=False):
            return True
        if self.ui.selection_items.object_display_count(key):
            self._set_background_on_display_row_under_pointer(key)
            return True
        if dodefaultaction:
            self._set_background_normal_row_under_pointer(key)
            return True
        return False

    def set_row(self, key, dodefaultaction=True, **kargs):
        """Return row widget for selection rule key or None."""
        row = super().set_row(key, dodefaultaction=False, **kargs)
        if row is not None:
            return row
        if key not in self.keys:
            return None
        if self.ui.selection_items.object_display_count(key):
            return self.objects[key].grid_row_on_display(**kargs)
        if dodefaultaction:
            return self.objects[key].grid_row_normal(**kargs)
        return None

    def launch_delete_record(self, key, modal=True):
        """Create delete dialogue."""
        oldobject = ChessDBrecordQuery()
        oldobject.set_database(self.ui.base_games.datasource.dbhome)
        oldobject.value.dbset = self.ui.base_games.datasource.dbset
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_delete_dialog(
            self.objects[key], oldobject, modal, title="Delete Selection Rule"
        )

    def launch_edit_record(self, key, modal=True):
        """Create edit dialogue."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordQuery(),
            ChessDBrecordQuery(),
            False,
            modal,
            title="Edit Selection Rule",
        )

    def launch_edit_show_record(self, key, modal=True):
        """Create edit dialogue including reference copy of original."""
        self.create_edit_dialog(
            self.objects[key],
            ChessDBrecordQuery(),
            ChessDBrecordQuery(),
            True,
            modal,
            title="Edit Selection Rule",
        )

    def launch_insert_new_record(self, modal=True):
        """Create insert dialogue."""
        instance = self.datasource.new_row()

        # Later process_query_statement() causes display of empty title and
        # query lines.
        instance.srvalue = repr("")

        self.create_edit_dialog(
            instance,
            ChessDBrecordQuery(),
            None,
            False,
            modal,
            title="New Selection Rule",
        )

    def launch_show_record(self, key, modal=True):
        """Create show dialogue."""
        oldobject = ChessDBrecordQuery()
        oldobject.set_database(self.ui.base_games.datasource.dbhome)
        oldobject.value.dbset = self.ui.base_games.datasource.dbset
        oldobject.load_record(
            (self.objects[key].key.pack(), self.objects[key].srvalue)
        )
        self.create_show_dialog(
            self.objects[key], oldobject, modal, title="Show Selection Rule"
        )

    def _set_grid_database(self, object_):
        object_.set_database(self.ui.base_games.datasource.dbhome)
        object_.value.dbset = self.ui.base_games.datasource.dbset

    def export_selection_rules(self, event=None):
        """Export selected selection rule definitions."""
        del event
        export_selection_rule.export_selected_selection_rules(
            self, self.ui.get_export_filename("Selection Rules", pgn=False)
        )


class QueryGrid(QueryListGrid):
    """Customized QueryListGrid for list of game selection rules."""

    def __init__(self, ui):
        """Extend with definition and bindings for selection rules on grid.

        ui - container for user interface widgets and methods.

        """
        super().__init__(ui.selection_rules_pw, ui)
        self.make_header(ChessDBrowQuery.header_specification)
        self.__bind_on()
        self._set_popup_bindings(
            self.menupopup,
            (
                (
                    EventSpec.display_record_from_grid,
                    self._display_selection_rule_from_popup,
                ),
                (
                    EventSpec.edit_record_from_grid,
                    self._edit_selection_rule_from_popup,
                ),
            ),
        )
        bindings = (
            (
                EventSpec.navigate_to_position_grid,
                self.set_focus_position_grid,
            ),
            (
                EventSpec.navigate_to_active_game,
                self._set_focus_gamepanel_item_command,
            ),
            (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
            (
                EventSpec.navigate_to_repertoire_grid,
                self.set_focus_repertoire_grid,
            ),
            (
                EventSpec.navigate_to_active_repertoire,
                self._set_focus_repertoirepanel_item_command,
            ),
            (
                EventSpec.navigate_to_repertoire_game_grid,
                self.set_focus_repertoire_game_grid,
            ),
            (
                EventSpec.navigate_to_active_partial,
                self._set_focus_partialpanel_item_command,
            ),
            (
                EventSpec.navigate_to_partial_game_grid,
                self.set_focus_partial_game_grid,
            ),
            (EventSpec.navigate_to_partial_grid, self.set_focus_partial_grid),
            (
                EventSpec.navigate_to_active_selection_rule,
                self._set_focus_selectionpanel_item_command,
            ),
            (EventSpec.tab_traverse_backward, self.traverse_backward),
            (EventSpec.tab_traverse_forward, self.traverse_forward),
        )
        self._add_cascade_menu_to_popup("Navigation", self.menupopup, bindings)
        self._add_cascade_menu_to_popup(
            "Navigation", self.menupopupnorow, bindings
        )

    def bind_off(self):
        """Disable all bindings."""
        super().bind_off()
        self._set_event_bindings_frame(
            (
                (EventSpec.navigate_to_active_partial, ""),
                (EventSpec.navigate_to_partial_game_grid, ""),
                (EventSpec.navigate_to_repertoire_grid, ""),
                (EventSpec.navigate_to_active_repertoire, ""),
                (EventSpec.navigate_to_repertoire_game_grid, ""),
                (EventSpec.navigate_to_position_grid, ""),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, ""),
                (EventSpec.navigate_to_partial_grid, ""),
                (EventSpec.navigate_to_active_selection_rule, ""),
                (EventSpec.display_record_from_grid, ""),
                (EventSpec.edit_record_from_grid, ""),
            )
        )

    def bind_on(self):
        """Enable all bindings."""
        super().bind_on()
        self.__bind_on()

    def __bind_on(self):
        """Enable all bindings."""
        self._set_event_bindings_frame(
            (
                (
                    EventSpec.navigate_to_active_partial,
                    self.set_focus_partialpanel_item,
                ),
                (
                    EventSpec.navigate_to_partial_game_grid,
                    self.set_focus_partial_game_grid,
                ),
                (
                    EventSpec.navigate_to_repertoire_grid,
                    self.set_focus_repertoire_grid,
                ),
                (
                    EventSpec.navigate_to_active_repertoire,
                    self.set_focus_repertoirepanel_item,
                ),
                (
                    EventSpec.navigate_to_repertoire_game_grid,
                    self.set_focus_repertoire_game_grid,
                ),
                (
                    EventSpec.navigate_to_position_grid,
                    self.set_focus_position_grid,
                ),
                (
                    EventSpec.navigate_to_active_game,
                    self.set_focus_gamepanel_item,
                ),
                (EventSpec.navigate_to_game_grid, self.set_focus_game_grid),
                (
                    EventSpec.navigate_to_partial_grid,
                    self.set_focus_partial_grid,
                ),
                (
                    EventSpec.navigate_to_active_selection_rule,
                    self.set_focus_selectionpanel_item,
                ),
                (
                    EventSpec.display_record_from_grid,
                    self._display_selection_rule,
                ),
                (EventSpec.edit_record_from_grid, self._edit_selection_rule),
            )
        )

    def _display_selection_rule(self, event=None):
        """Display selection rule and cancel selection.

        Call _display_selection_rule_after_idle after idle tasks to allow
        message display.

        """
        del event
        if not self.get_visible_selected_key():
            return
        self._set_find_selection_rule_name_games(self.selection[0])
        self.frame.after_idle(
            self.try_command(
                self._display_selection_rule_after_idle, self.frame
            )
        )

    def _display_selection_rule_from_popup(self, event=None):
        """Display selection rule selected by pointer.

        Call _display_selection_rule_after_idle after idle tasks to allow
        message display.

        """
        del event
        self._set_find_selection_rule_name_games(self.pointer_popup_selection)
        self.frame.after_idle(
            self.try_command(
                self._display_selection_rule_from_popup_after_idle, self.frame
            )
        )

    def _display_selection_rule_after_idle(self):
        """Display selection rule and cancel selection.

        Call from _display_selection_rule only.

        """
        self._display_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def _display_selection_rule_from_popup_after_idle(self):
        """Display selection rule selected by pointer.

        Call from _display_selection_rule_from_popup only.

        """
        self._display_selected_item(self.pointer_popup_selection)

    def _edit_selection_rule(self, event=None):
        """Display selection rule allow editing and cancel selection.

        Call _edit_selection_rule_after_idle after idle tasks to allow
        message display.

        """
        del event
        if not self.get_visible_selected_key():
            return
        self._set_find_selection_rule_name_games(self.selection[0])
        self.frame.after_idle(
            self.try_command(self._edit_selection_rule_after_idle, self.frame)
        )

    def _edit_selection_rule_from_popup(self, event=None):
        """Display selection rule with editing allowed selected by pointer.

        Call _edit_selection_rule_after_idle after idle tasks to allow
        message display.

        """
        del event
        self._set_find_selection_rule_name_games(self.pointer_popup_selection)
        self.frame.after_idle(
            self.try_command(
                self._edit_selection_rule_from_popup_after_idle, self.frame
            )
        )

    def _edit_selection_rule_after_idle(self):
        """Display selection rule allow editing and cancel selection.

        Call from _edit_selection_rule only.

        """
        self._edit_selected_item(self.get_visible_selected_key())
        self.cancel_selection()

    def _edit_selection_rule_from_popup_after_idle(self):
        """Display selection rule with editing allowed selected by pointer.

        Call from _edit_selection_rule_from_popup only.

        """
        self._edit_selected_item(self.pointer_popup_selection)

    def _set_find_selection_rule_name_games(self, key):
        """Set status text to active selection rule name."""
        if self.ui.selection_items.count_items_in_stack():
            # do search at this time only if no selection rules displayed
            return
        self.ui.statusbar.set_status_text(
            "".join(
                (
                    "Please wait while finding games for selection rule ",
                    self.objects[key].value.get_name_text(),
                )
            )
        )

    def on_partial_change(self, instance):
        # may turn out to be just to catch datasource is None
        """Refresh query grid after database update for instance."""
        if self.get_data_source() is None:
            return
        super().on_data_change(instance)

    def set_selection_text(self):
        """Set status bar to display selection rule name."""
        if self.selection:
            value = self.objects[self.selection[0]].value
            self.ui.statusbar.set_status_text(
                "".join(
                    (
                        value.get_name_text(),
                        "   (",
                        value.get_query_statement_text(),
                        ")",
                    )
                )
            )
        else:
            self.ui.statusbar.set_status_text("")

    def is_visible(self):
        """Return True if list of selection rules is displayed."""
        return str(self.get_frame()) in self.ui.selection_rules_pw.panes()

    def make_display_widget(self, sourceobject):
        """Return a QueryDisplay for sourceobject."""
        selection = super().make_display_widget(sourceobject)
        selection.set_and_tag_item_text()
        return selection

    def make_edit_widget(self, sourceobject):
        """Return a QueryDisplayEdit for sourceobject."""
        selection = super().make_edit_widget(sourceobject)
        selection.set_and_tag_item_text(reset_undo=True)
        return selection

    def focus_set_frame(self, event=None):
        """Delegate to superclass then set toolbar widget states."""
        super().focus_set_frame(event=event)
        self.ui.set_toolbarframe_normal(
            self.ui.move_to_selection, self.ui.filter_selection
        )

    def set_selection(self, key):
        """Hack to fix edge case when inserting records using apsw or sqlite3.

        Workaround a KeyError exception when a record is inserted while a grid
        keyed by a secondary index with only one key value in the index is on
        display.

        """
        try:
            super().set_selection(key)
        except KeyError:
            tkinter.messagebox.showinfo(
                parent=self.parent,
                title="Insert Selection Rule Workaround",
                message="".join(
                    (
                        "All records have same name on this display.\n\nThe ",
                        "new record has been inserted but you need to Hide, ",
                        "and then Show, the display to see the record in ",
                        "the list.",
                    )
                ),
            )



# cqldbedit.py
# Copyright 2016 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Customise edit toplevel to edit or insert ChessQL statement record.

ChessQL statements obey the syntax published for CQL version 6.0.1 (by Gady
Costeff).

"""
import tkinter.messagebox

from solentware_grid.gui.dataedit import DataEdit

from .cqltoplevel import CQLToplevel, CQLToplevelEdit
from .topleveltext import EditText


class CQLDbEdit(EditText, DataEdit):
    """Edit ChessQL statement on database, or insert a new record.

    parent is used as the master argument in CQLToplevel calls.

    ui is used as the ui argument in CQLToplevel calls.

    newobject, parent, oldobject, and the one or two CQLToplevel instances
    created, are used as arguments in the super.__init__ call.

    showinitial determines whether a CQLToplevel is created for oldobject if
    there is one.

    Attribute text_name provides the name used in widget titles and message
    text.

    Methods _get_title_for_object and _set_item, and properties ui_base_table;
    ui_items_in_toplevels; and ui, allow similar methods in various classes
    to be expressed identically and defined once.

    """

    text_name = "ChessQL Statement"

    def __init__(
        self,
        newobject=None,
        parent=None,
        oldobject=None,
        showinitial=True,
        ui=None,
    ):
        """Extend and create toplevel to edit or insert ChessQL statement."""
        if not oldobject:
            showinitial = False
        super().__init__(
            newobject=newobject,
            parent=parent,
            oldobject=oldobject,
            newview=CQLToplevelEdit(master=parent, ui=ui),
            title="",
            oldview=CQLToplevel(master=parent, ui=ui)
            if showinitial
            else showinitial,
        )
        self._initialize()

    def _get_title_for_object(self, object_=None):
        """Return title for Toplevel containing a ChessQL statement object_.

        Default value of object_ is oldobject attribute from DataEdit class.

        """
        if object_ is None:
            object_ = self.oldobject
        if object_:
            return "  ".join(
                (
                    self.text_name.join(("Edit ", ":")),
                    object_.value.get_name_text(),
                )
            )
        return "".join(("Insert ", self.text_name))

    @property
    def ui_base_table(self):
        """Return the User Interface TagRosterGrid object."""
        return self.ui.base_partials

    @property
    def ui_items_in_toplevels(self):
        """Return the User Interface objects in Toplevels."""
        return self.ui.partials_in_toplevels

    @property
    def ui(self):
        """Return the User Interface object from 'editable' view."""
        return self.newview.ui

    @staticmethod
    def _set_item(view, object_):
        """Populate view with the CQL query extracted from object_."""
        view.cql_statement.process_statement(object_.get_srvalue())
        view.set_and_tag_item_text(reset_undo=True)

    def dialog_ok(self):
        """Return update action response (True for deleted).

        Delegate to superclass if the active item on the main display is not a
        'Position | Partial' widget and contains a valid statement.

        """
        self.newobject.value.load(
            repr(self.newview.get_name_cql_statement_text())
        )
        title = self._get_title_for_object()
        if not self.newobject.value.get_name_text():
            if not self.newobject.value.cql_error:
                tkinter.messagebox.showerror(
                    parent=self.parent,
                    title=title,
                    message="".join(
                        (
                            "The ChessQL statement has no name.\n\nPlease ",
                            "enter it's name as the first line of text.'",
                        )
                    ),
                )
            else:
                tkinter.messagebox.showerror(
                    parent=self.parent,
                    title=title,
                    message="".join(
                        (
                            "The text does not contain a valid ChessQL ",
                            "statement. ",
                        )
                    ),
                )
            return False
        if self.newobject.value.cql_error:
            if tkinter.messagebox.YES != tkinter.messagebox.askquestion(
                parent=self.parent,
                title=title,
                message="".join(
                    (
                        "Confirm request to update ChessQL statement ",
                        "named:\n\n",
                        self.newobject.value.get_name_text(),
                        "\n\non database.\n\n",
                        self.newobject.value.cql_error.get_error_report(),
                    )
                ),
            ):
                return False
        if self.ui.partial_items.active_item:
            if self.ui.partial_items.active_item.sourceobject is None:
                tkinter.messagebox.showinfo(
                    parent=self.parent,
                    title=title,
                    message="".join(
                        (
                            "Cannot use this insert dialogue while the ",
                            "active item in cql queries is one opened by ",
                            "menu action 'Position | Partial'.",
                        )
                    ),
                )
                return False
        return super().dialog_ok()

    def edit(self, commit=True):
        """Delegate to superclass to edit record then update game list."""
        if commit:
            self.datasource.dbhome.start_transaction()
        # Hack to prevent _lmdb interface via lmdb to Symas LMMD crash.
        # Do a widget navigation action to cause refresh after edit in lmdb.
        if self.datasource.dbhome.dbenv.__class__.__name__ == "Environment":
            super().edit_no_refresh(commit=False)
        else:
            super().edit(commit=False)
        cqls = self.ui.partialpositionds(
            self.ui.base_games.datasource.dbhome,
            self.ui.base_games.datasource.dbset,
            self.ui.base_games.datasource.dbset,
            newrow=None,
        )
        cqls.update_cql_statement_games(self.newobject, commit=False)
        if commit:
            self.datasource.dbhome.commit()

    def put(self, commit=True):
        """Delegate to superclass to insert record then insert game list."""
        if commit:
            self.datasource.dbhome.start_transaction()
        super().put(commit=False)
        cqls = self.ui.partialpositionds(
            self.ui.base_games.datasource.dbhome,
            self.ui.base_games.datasource.dbset,
            self.ui.base_games.datasource.dbset,
            newrow=None,
        )
        cqls.update_cql_statement_games(self.newobject, commit=False)
        if commit:
            self.datasource.dbhome.commit()



# dbtkinterdu_dir.py
# Copyright 2023 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Import directory of PGN files with db_tkinter.chessdbtkinterdu module."""


if __name__ == "__main__":

    from .directory_widget import DirectoryWidget
    from ..db_tkinter.chessdbtkinterdu import chess_database_du

    DirectoryWidget(chess_database_du, "db_tcl")



# runchessdptdu.py
# Copyright 2009 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess database update using DPT single-step deferred update.

Run as a new process from the chess GUI.

"""

if __name__ == "__main__":

    # run by subprocess.popen from ../core/chess.py
    import sys
    import os

    try:
        # If module not loaded from Python site-packages put the folder
        # containing chesstab at front of sys.path on the assumption all the
        # sibling packages are there too.
        try:
            sp = sys.path[-1].replace("\\\\", "\\")
            packageroot = os.path.dirname(
                os.path.dirname(os.path.dirname(__file__))
            )
            if sp != packageroot:
                sys.path.insert(0, packageroot)
        except NameError as msg:
            # When run in the py2exe generated executable the module will
            # not have the __file__ attribute.
            # But the siblings can be assumed to be in the right place.
            if " '__file__' " not in str(msg):
                raise

        # sys.path should now contain correct chesstab modules
        from chesstab.dpt import chessdptdu
        from chesstab.gui import chessdu

        cdu = chessdu.ChessDeferredUpdate(
            deferred_update_method=chessdptdu.chess_dptdu,
            database_class=chessdptdu.ChessDatabase,
        )
    except Exception as error:
        try:
            chessdu.write_error_to_log()
        except Exception:
            # Assume that parent process will report the failure
            raise SystemExit(
                " reporting exception in ".join(("Exception while", "DPTdu"))
            ) from error
        raise SystemExit("Reporting exception in DPTdu") from error



# chessdpt.py
# Copyright 2008 Roger Marsh
# Licence: See LICENCE (BSD licence)

"""Chess games database using DPT database via dptdb.dptapi."""

import os
import tkinter.messagebox

# pylint will always give import-error message on non-Microsoft Windows
# systems.
# Wine counts as a Microsft Windows system.
# It is reasonable to not install 'dptdb.dptapi'.
# The importlib module is used to import chessdpt if needed.
from dptdb.dptapi import (
    FIFLAGS_FULL_TABLEB,
    FIFLAGS_FULL_TABLED,
    FISTAT_DEFERRED_UPDATES,
)

from solentware_base import dpt_database
from solentware_base.core.constants import FILEDESC

from ..core.filespec import FileSpec
from ..basecore import database
from .. import APPLICATION_NAME


class ChessdptError(Exception):
    """Exception class for chessdpt module."""


class ChessDatabase(database.Database, dpt_database.Database):
    """Provide access to a database of games of chess."""

    # The default for DPT.  See use_deferred_update_process method for cases
    # where this is not used: normally delegated to superclass to pick the
    # default.
    _deferred_update_process = os.path.join(
        os.path.basename(os.path.dirname(__file__)), "runchessdptdu.py"
    )

    def __init__(
        self,
        databasefolder,
        use_specification_items=None,
        dpt_records=None,
        **kargs,
    ):
        """Define chess database.

        **kargs
        allowcreate == False - remove file descriptions from FileSpec so
        that superclass cannot create them.
        Other arguments are passed through to superclass __init__.

        """
        try:
            sysprint = kargs.pop("sysprint")
        except KeyError:
            sysprint = "CONSOLE"
        ddnames = FileSpec(
            use_specification_items=use_specification_items,
            dpt_records=dpt_records,
        )

        if not kargs.get("allowcreate", False):
            try:
                for dd_name in ddnames:
                    if FILEDESC in ddnames[dd_name]:
                        del ddnames[dd_name][FILEDESC]
            except Exception as error:
                if __name__ == "__main__":
                    raise
                raise ChessdptError("DPT description invalid") from error

        try:
            super().__init__(
                ddnames, databasefolder, sysprint=sysprint, **kargs
            )
        except ChessdptError as error:
            if __name__ == "__main__":
                raise
            raise ChessdptError("DPT description invalid") from error

        self._broken_sizes = dict()

    def use_deferred_update_process(self, **kargs):
        """Return path to deferred update module."""
        chunk = self._use_deferred_update_process_chunk(**kargs)
        if chunk is not None:
            return chunk
        return super().use_deferred_update_process(**kargs)

    @staticmethod
    def _use_deferred_update_process_chunk(
        dptmultistepdu=False, dptchunksize=None, **kargs
    ):
        """Return module name or None.

        dptmultistepdu is ignored if dptchunksize is not None.
        dptmultistepdu is ignored because multi-step is no longer supported.

        dptchunksize is None: dptmultistepdu determines deferred update module
        otherwise use single-step deferred update with the chunk size (assumed
        to be a valid chunk size)

        **kargs - soak up any arguments other database engines need.

        On non-Microsoft operating systems single-step update may not work.
        First encountered on upgrade to FreeBSD7.2 wine-1.1.23,1 dptv2r19.
        But single-step works on FreeBSD7.2 wine-1.1.0,1 dptv2r19.
        Wine source code comments state that memory-use calls do not return
        correct values. Evidently the incorrectness can vary by version
        making the DPT workarounds liable to fail also.

        runchessdptdu.py does DPT's single-step deferred update process.
        runchessdptduchunk.py does DPT's single-step deferred update process
        but splits the task into fixed size chunks, a number of games, which
        it is hoped are small enough to finish before all memory is used.
        The deleted runchessdptdumulti.py did DPT's multi-step deferred
        update process.

        Multi-step was about half an order of magnitude slower than
        single-step.

        """
        del dptmultistepdu, kargs
        if dptchunksize is not None:
            return os.path.join(
                os.path.basename(os.path.dirname(__file__)),
                "runchessdptduchunk.py",
            )
        return None

    def adjust_database_for_retry_import(self, files):
        """Increase file sizes taking file full into account."""
        # Increase the size of files allowing for the file full condition
        # which occurred while doing a deferred update for import.
        for dbn in self._broken_sizes:
            self.table[dbn].increase_size_of_full_file(
                self.dbenv,
                self.table[dbn].get_file_parameters(self.dbenv),
                self._broken_sizes[dbn],
            )

    def open_database(self, files=None):
        """Return True if all files are opened in Normal mode (FISTAT == 0)."""
        super().open_database(files=files)
        fistat = dict()
        for dbo in self.table.values():
            fistat[dbo] = dbo.get_file_parameters(self.dbenv)["FISTAT"]
        for dbo in self.table.values():
            if fistat[dbo][0] != 0:
                break
        else:
            self.increase_database_size(files=None)
            return True

        # At least one file is not in Normal state
        report = "\n".join(
            [
                "\t".join((os.path.basename(dbo.file), fistat[dbo][1]))
                for dbo in self.table.values()
            ]
        )
        tkinter.messagebox.showinfo(
            title="Open",
            message="".join(
                (
                    APPLICATION_NAME,
                    " has opened the database but some of the files are ",
                    "not in the Normal state.\n\n",
                    report,
                    "\n\n",
                    APPLICATION_NAME,
                    " will close the database on dismissing this ",
                    "dialogue.\n\nRestore the database from backups, or ",
                    "source data, before trying again.",
                )
            ),
        )
        self.close_database()

    def _delete_database_names(self):
        """Override and return tuple of filenames to delete."""
        names = [self.sysfolder]
        for value in self.table.values():
            names.append(value.file)
        return tuple(names)

    def get_archive_names(self, files=()):
        """Return names and operating system files for archives and guards."""
        specs = {f for f in files if f in self.table}
        names = [v.file for k, v in self.table.items() if k in specs]
        archives = dict()
        guards = dict()
        for name in names:
            archiveguard = ".".join((name, "grd"))
            archivefile = ".".join((name, "bz2"))
            for box, file in ((archives, archivefile), (guards, archiveguard)):
                if os.path.exists(file):
                    box[name] = file
        return (names, archives, guards)

    def open_after_import_without_backups(self, files=()):
        """Return open context after doing database engine specific actions.

        For DPT clear the file sizes before import area if the database was
        opened successfully as there is no need to retry the import.

        """
        super().open_database()
        fistat = dict()
        file_sizes_for_import = dict()
        for dbn, dbo in self.table.items():
            gfp = dbo.get_file_parameters(self.dbenv)
            fistat[dbo] = gfp["FISTAT"]
            if dbn in files:
                file_sizes_for_import[dbn] = gfp
        for dbo in self.table.values():
            if fistat[dbo][0] != 0:
                break
        else:
            # Assume all is well as file status is 0
            # Or just do nothing (as file_sizes_for_import may be removed)
            self.increase_database_size(files=None)
            self.mark_partial_positions_to_be_recalculated()
            return True
        # At least one file is not in Normal state after Import.
        # Check the files that had imports applied
        for dbn in file_sizes_for_import:
            # pylint message unused variable.
            # Document what seemed to matter at some point.
            # status = file_sizes_for_import[dbn]["FISTAT"][0]
            flags = file_sizes_for_import[dbn]["FIFLAGS"]
            if not (
                (flags & FIFLAGS_FULL_TABLEB) or (flags & FIFLAGS_FULL_TABLED)
            ):
                break
        else:
            # The file states are consistent with the possibility that the
            # import failed because at least one file was too small.
            # The file size information is kept for calculating an increase
            # in file size before trying the import again.
            tkinter.messagebox.showinfo(
                title="Open",
                message="".join(
                    (
                        "The import failed.\n\n",
                        APPLICATION_NAME,
                        " has opened the database but some of the files are ",
                        "full and backups were not taken, so cannot offer ",
                        "the option of retrying the import with a larger ",
                        "file, and cannot restore the database.  The ",
                        "database may not be usable.",
                    )
                ),
            )
            self.close_database()
            return None
        # At least one file is not in Normal state.
        # None of these files had deferred updates for Import or the state does
        # not imply a file full condition where deferred updates occured.
        report = "\n".join(
            [
                "\t".join((os.path.basename(dbo.file), fistat[dbo][1]))
                for dbo in self.table.values()
            ]
        )
        action = tkinter.messagebox.askyesno(
            title="Open",
            message="".join(
                (
                    APPLICATION_NAME,
                    " has opened the database but some of the files are ",
                    "not in the Normal state.\n\n",
                    report,
                    "\n\nAt least one of these files is neither just ",
                    "marked Deferred Update nor marked Full, and backups ",
                    "were not taken, so ",
                    APPLICATION_NAME,
                    " is not offering the option of ",
                    "retrying the import with a larger file.\n\nDo you ",
                    "want to save a copy of the broken database?",
                )
            ),
        )
        self.close_database()
        if not action:
            return "Import failed"
        return False

    def open_after_import_with_backups(self, files=()):
        """Return open context after doing database engine specific actions.

        For DPT clear the file sizes before import area if the database was
        opened successfully as there is no need to retry the import.

        """
        super().open_database()

        # open_database() call after completion of Import sequence
        fistat = dict()
        file_sizes_for_import = dict()
        for dbn, dbo in self.table.items():
            gfp = dbo.get_file_parameters(self.dbenv)
            fistat[dbo] = gfp["FISTAT"]
            if dbn in files:
                file_sizes_for_import[dbn] = gfp
        for dbo in self.table.values():
            if fistat[dbo][0] != 0:
                break
        else:
            self.increase_database_size(files=None)
            self.mark_partial_positions_to_be_recalculated()
            return True
        # At least one file is not in Normal state after Import.
        # Check the files that had imports applied
        for dbn in file_sizes_for_import:
            status = file_sizes_for_import[dbn]["FISTAT"][0]
            flags = file_sizes_for_import[dbn]["FIFLAGS"]
            if not (
                (status == 0)
                or (status == FISTAT_DEFERRED_UPDATES)
                or (flags & FIFLAGS_FULL_TABLEB)
                or (flags & FIFLAGS_FULL_TABLED)
            ):
                break
        else:
            # The file states are consistent with the possibility that the
            # import failed because at least one file was too small.
            # The file size information is kept for calculating an increase
            # in file size before trying the import again.
            if tkinter.messagebox.askyesno(
                title="Retry Import",
                message="".join(
                    (
                        "The import failed because the games file was filled.",
                        "\n\nThe file will be restored from backups.\n\nDo ",
                        "you want to retry the import with more space (20%) ",
                        "allocated to the games file?",
                    )
                ),
            ):
                return None
            self.close_database()
            return "Restore without retry"
        # At least one file is not in Normal state.
        # None of these files had deferred updates for Import or the state does
        # not imply a file full condition where deferred updates occured.
        report = "\n".join(
            [
                "\t".join((os.path.basename(dbo.file), fistat[dbo][1]))
                for dbo in self.table.values()
            ]
        )
        action = tkinter.messagebox.askyesno(
            title="Open",
            message="".join(
                (
                    APPLICATION_NAME,
                    " has opened the database but some of the files are ",
                    "not in the Normal state.\n\n",
                    report,
                    "\n\nAt least one of these files is neither just ",
                    "marked Deferred Update nor marked Full so ",
                    APPLICATION_NAME,
                    " is not offering the option of retrying ",
                    "the import with a larger file.\n\nDo you want to save a ",
                    "copy of the broken database before restoring from ",
                    "backups?",
                )
            ),
        )
        self.close_database()
        if not action:
            return "Import failed"
        return False

    def save_broken_database_details(self, files=()):
        """Save database engine specific detail of broken files to be restored.

        It is assumed that the Database Services object exists.

        """
        self._broken_sizes.clear()
        broken = self._broken_sizes
        for file in files:
            broken[file] = self.table[file].get_file_parameters(self.dbenv)
